# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.lib.common import supplement_filesystem
from storops.exception import UnityResourceNotFoundError, \
    UnityCifsServiceNotEnabledError, UnityShareShrinkSizeTooLargeError, \
    UnityShareShrinkSizeTooSmallError, UnityLocalReplicationFsNameNotSameError
from storops.unity.enums import FSSupportedProtocolEnum, TieringPolicyEnum, \
    SnapStateEnum

import storops.unity.resource.nas_server
import storops.unity.resource.pool
import storops.unity.resource.nfs_share
import storops.unity.resource.cifs_share
from storops.unity.resource import UnityResource, UnityResourceList
from storops.unity.resource.replication_session import \
    UnityReplicationSession, UnityResourceConfig
from storops.unity.resource.snap import UnitySnap, UnitySnapList
from storops.unity.resource.storage_resource import UnityStorageResource

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


class UnityFileSystem(UnityResource):
    @classmethod
    def create(cls, cli, pool, nas_server, name, size,
               proto=None, is_thin=None,
               tiering_policy=None, user_cap=False):
        pool_clz = storops.unity.resource.pool.UnityPool
        nas_server_clz = storops.unity.resource.nas_server.UnityNasServer

        if proto is None:
            proto = FSSupportedProtocolEnum.NFS

        pool = pool_clz.get(cli, pool)
        nas_server = nas_server_clz.get(cli, nas_server)
        FSSupportedProtocolEnum.verify(proto)
        TieringPolicyEnum.verify(tiering_policy)
        size = supplement_filesystem(size, user_cap)

        req_body = {
            'name': name,
            'fsParameters': {
                'pool': pool,
                'nasServer': nas_server,
                'supportedProtocols': proto,
                'isThinEnabled': is_thin,
                'size': size,
                'fastVPParameters': {
                    'tieringPolicy': tiering_policy
                }
            },
        }
        resp = cli.type_action(UnityStorageResource().resource_class,
                               'createFilesystem',
                               **req_body)
        resp.raise_if_err()
        sr = UnityStorageResource(_id=resp.resource_id, cli=cli)
        return sr.filesystem

    @property
    def first_available_cifs_server(self):
        ret = None
        if self.nas_server is not None:
            try:
                ret = self.nas_server.get_cifs_server()
            except UnityCifsServiceNotEnabledError as e:
                log.info(e.message)
        return ret

    def delete(self, force_snap_delete=False, force_vvol_delete=False,
               async_mode=False):
        sr = self.storage_resource
        if not self.existed or sr is None:
            raise UnityResourceNotFoundError(
                'cannot find filesystem {}.'.format(self.get_id()))
        resp = self._cli.delete(sr.resource_class,
                                sr.get_id(),
                                forceSnapDeletion=force_snap_delete,
                                forceVvolDeletion=force_vvol_delete,
                                async_mode=async_mode)
        resp.raise_if_err()
        return resp

    def extend(self, new_size, user_cap=False):
        sr = self.storage_resource
        new_size = supplement_filesystem(new_size, user_cap)
        param = self._cli.make_body(size=new_size)
        resp = sr.modify_fs(fsParameters=param)
        resp.raise_if_err()
        return resp

    def shrink(self, new_size, user_cap=False):
        sr = self.storage_resource
        new_size = supplement_filesystem(new_size, user_cap)
        size_used = sr.size_used
        if size_used and int(size_used) > new_size:
            message = 'Reject shrink share request, ' \
                      'the new size should be larger than used.'
            raise UnityShareShrinkSizeTooSmallError(message)
        param = self._cli.make_body(size=new_size)
        size_total = sr.size_total
        if size_total and int(size_total) < new_size:
            message = 'Reject shrink share request, ' \
                      'the new size should be smaller than original.'
            raise UnityShareShrinkSizeTooLargeError(message)
        resp = sr.modify_fs(fsParameters=param)
        resp.raise_if_err()
        return resp

    def create_nfs_share(self, name, path=None, share_access=None):
        clz = storops.unity.resource.nfs_share.UnityNfsShare
        return clz.create(self._cli, name=name, fs=self,
                          path=path, share_access=share_access)

    def create_cifs_share(self, name, path=None, cifs_server=None):
        clz = storops.unity.resource.cifs_share.UnityCifsShare
        return clz.create(self._cli, name=name, fs=self,
                          path=path, cifs_server=cifs_server)

    def create_snap(self, name=None,
                    description=None, is_auto_delete=None,
                    retention_duration=None, is_read_only=None,
                    fs_access_type=None):
        return UnitySnap.create(cli=self._cli,
                                storage_resource=self.storage_resource,
                                name=name,
                                description=description,
                                is_auto_delete=is_auto_delete,
                                retention_duration=retention_duration,
                                is_read_only=is_read_only,
                                fs_access_type=fs_access_type)

    @property
    def snapshots(self):
        return UnitySnapList(cli=self._cli,
                             storage_resource=self.storage_resource)

    def has_snap(self, ignore_system_snap=False):
        """ This method won't count the snaps in "destroying" state!

        :param ignore_system_snap: ignore the system snap if True.
        :return: false if no snaps or all snaps are destroying.
        """
        snaps = filter(lambda s: s.state != SnapStateEnum.DESTROYING,
                       self.snapshots)
        if ignore_system_snap:
            snaps = filter(lambda s: not s.is_system_snap, snaps)
        return len(list(snaps)) > 0

    def replicate_with_dst_resource_provisioning(self, max_time_out_of_sync,
                                                 dst_pool_id,
                                                 dst_fs_name=None,
                                                 remote_system=None,
                                                 replication_name=None,
                                                 dst_size=None,
                                                 is_dst_thin=None,
                                                 dst_tiering_policy=None,
                                                 is_dst_compression=None):
        """
        Creates a replication session with destination filesystem provisioning.

        :param max_time_out_of_sync: maximum time to wait before syncing the
            source and destination. Value `-1` means the automatic sync is not
            performed. `0` means it is a sync replication.
        :param dst_pool_id: id of pool to allocate destination filesystem.
        :param dst_fs_name: destination filesystem name. If `remote_system` is
            `None` (for local replication creation), `dst_fs_name` should be
            same as the source fs name or `None`.
        :param remote_system: `UnityRemoteSystem` object. The remote system to
            which the replication is being configured. When not specified, it
            defaults to local system.
        :param replication_name: replication name.
        :param dst_size: destination filesystem size.
        :param is_dst_thin: indicates whether destination filesystem is thin or
            not.
        :param dst_tiering_policy: `TieringPolicyEnum` value. Tiering policy of
            destination filesystem.
        :param is_dst_compression: indicates whether destination filesystem is
            compression enabled or not.
        :return: created replication session.
        """

        if dst_fs_name is None:
            dst_fs_name = self.name
        if remote_system is None and dst_fs_name != self.name:
            raise UnityLocalReplicationFsNameNotSameError(
                'dst_fs_name passed in for creating filesystem local '
                'replication should be same as source filesystem name '
                'or None')

        dst_size = self.size_total if dst_size is None else dst_size

        dst_resource = UnityResourceConfig.to_embedded(
            name=dst_fs_name, pool_id=dst_pool_id, size=dst_size,
            tiering_policy=dst_tiering_policy, is_thin_enabled=is_dst_thin,
            is_compression_enabled=is_dst_compression)
        return UnityReplicationSession.create_with_dst_resource_provisioning(
            self._cli, self.storage_resource.get_id(),
            dst_resource, max_time_out_of_sync,
            remote_system=remote_system, name=replication_name)


class UnityFileSystemList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityFileSystem



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import functools
import json
import logging
import os

from mock import patch

from storops.exception import MockFileNotFoundError
from storops.lib.common import cache, allow_omit_parentheses
from storops.unity.client import UnityClient
import storops.unity.resource.system
from storops_test.utils import ConnectorMock, read_test_file

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


@cache
def t_rest(version=None):
    """ get the test unity client

    :return: unity client singleton
    """
    client = UnityClient('10.244.223.61', 'admin', 'Password123!',
                         verify=False)
    client.set_system_version(version)
    return client


@cache
def t_unity(version=None):
    clz = storops.unity.resource.system.UnitySystem
    unity = clz('10.244.223.61', 'admin', 'Password123!', verify=False)
    unity.add_metric_record(unity.get_metric_query_result(17))
    unity.add_metric_record(unity.get_metric_query_result(34))
    unity._cli.set_system_version(version)
    return unity


class MockRestClient(ConnectorMock):
    base_folder = os.path.join('unity', 'rest_data')

    @classmethod
    def get_folder_from_url(cls, url):
        try:
            pos = url.index('?')
            url = url[:pos]
        except ValueError:
            pass
        if len(url.strip('/').split('/')) >= 3:
            return url.strip('/').split('/')[2]
        return ""

    def get(self, url, **kwargs):
        return self._get_mock_output(url, kwargs)

    def post(self, url, **kwargs):
        return self._get_mock_output(url, kwargs)

    def delete(self, url, **kwargs):
        return self._get_mock_output(url, kwargs)

    @classmethod
    def get_folder(cls, inputs):
        url, body = inputs
        return os.path.join(cls.base_folder, cls.get_folder_from_url(url))

    @classmethod
    def get_filename(cls, inputs):
        url, body = inputs
        indices = cls.read_index(cls.get_folder(inputs))
        for index in indices.get('indices', []):
            if url.lower() != index['url'].lower():
                continue
            elif not cls.compare_json_body(body, index.get('body', None)):
                continue
            response = index['response']
            break
        else:
            raise MockFileNotFoundError('cannot find response for url: {}, \n'
                                        'body: {}'.format(url, body))
        return response

    @staticmethod
    @cache
    def read_index(folder):
        string_indices = read_test_file(folder, 'index.json')
        return json.loads(string_indices, encoding='utf-8')

    def _get_mock_output(self, url, kwargs):
        body = kwargs.get('body', None)
        resp_body = self.get_mock_output([url, body])
        if len(resp_body) > 0:
            ret = json.loads(resp_body)
        else:
            ret = None
        return ret

    @classmethod
    def compare_json_body(cls, obj1, obj2):
        if all((not obj1, not obj2)):
            # cover the case: obj1={}, obj2=None
            return True
        else:
            return cls.ordered(obj1) == cls.ordered(obj2)

    @classmethod
    def ordered(cls, obj):
        """Normalize the JSON object.

        from https://stackoverflow.com/questions/25851183/
        how-to-compare-two-json-objects-with-the-
        same-elements-in-a-different-order-equa
        """
        if isinstance(obj, dict):
            return sorted((k, cls.ordered(v)) for k, v in obj.items())
        if isinstance(obj, list):
            return sorted(cls.ordered(x) for x in obj)
        else:
            return obj


@allow_omit_parentheses
def patch_rest(output=None, mock_map=None):
    rest = MockRestClient(output, mock_map)

    def decorator(func):
        @functools.wraps(func)
        @patch(target='storops.connection.connector.UnityRESTConnector.get',
               new=rest.get)
        @patch(target='storops.connection.connector.UnityRESTConnector.post',
               new=rest.post)
        @patch(target='storops.connection.connector.UnityRESTConnector.delete',
               new=rest.delete)
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return func_wrapper

    return decorator



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, raises

from storops.vnx.resource import VNXCliResource

__author__ = 'Cedric Zhuang'


class VNXCliResourceTest(TestCase):
    def test_with_poll(self):
        r = VNXCliResource()
        assert_that(r.poll, equal_to(True))
        r.poll = False
        assert_that(r.poll, equal_to(False))
        with r.with_poll():
            assert_that(r.poll, equal_to(True))
        assert_that(r.poll, equal_to(False))

    def test_no_with(self):
        r = VNXCliResource()
        r.with_poll()
        assert_that(r.poll, equal_to(True))
        r.with_no_poll()
        assert_that(r.poll, equal_to(False))

    def test_with_no_poll(self):
        r = VNXCliResource()
        assert_that(r.poll, equal_to(True))
        with r.with_no_poll():
            assert_that(r.poll, equal_to(False))
        assert_that(r.poll, equal_to(True))

    class DemoCliResourceRaiseException(VNXCliResource):
        @staticmethod
        def do():
            raise ValueError('test error.')

    def test_exception_in_with(self):
        def f():
            r = self.DemoCliResourceRaiseException()
            with r.with_no_poll():
                r.do()
        assert_that(f, raises(ValueError, 'test error'))



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# coding=utf-8
from __future__ import unicode_literals

from storops.unity.resource import UnityResource, UnityResourceList

__author__ = 'Cedric Zhuang'


class UnityDisk(UnityResource):
    @classmethod
    def get_nested_properties(cls):
        return (
            'pool.name',
        )

    @property
    def inserted(self):
        return self.raw_size > 0


class UnityDiskList(UnityResourceList):
    def __init__(self, **the_filter):
        if 'inserted' in the_filter:
            self._inserted = the_filter['inserted']
            del the_filter['inserted']
        else:
            self._inserted = None

        super(UnityDiskList, self).__init__(**the_filter)

    def _filter(self, item):
        ret = True
        if self._inserted is not None:
            ret = item.inserted == self._inserted
        return ret

    @classmethod
    def get_resource_class(cls):
        return UnityDisk


class UnityDiskGroup(UnityResource):
    pass


class UnityDiskGroupList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityDiskGroup



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.exception import raise_if_err, VNXSecurityException
from storops.vnx.resource import VNXCliResource, VNXCliResourceList

__author__ = 'Cedric Zhuang'


class VNXBlockUser(VNXCliResource):
    def __init__(self, name=None, cli=None):
        super(VNXBlockUser, self).__init__()
        self._cli = cli
        self._name = name

    @classmethod
    def get(cls, cli, name=None):
        if name is None:
            ret = VNXBlockUserList(cli=cli)
        else:
            ret = VNXBlockUser(name=name, cli=cli)
        return ret

    @classmethod
    def create(cls, cli, name, password, scope=None, role=None):
        out = cli.add_user(name, password, scope=scope, role=role)
        raise_if_err(out, default=VNXSecurityException)
        return VNXBlockUser(name=name, cli=cli)

    def delete(self):
        out = self._cli.delete_user(name=self._get_name(), scope=self.scope)
        raise_if_err(out, default=VNXSecurityException)

    def _get_raw_resource(self):
        return self._cli.list_user(name=self._name)


class VNXBlockUserList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXBlockUser

    def _get_raw_resource(self):
        return self._cli.list_user()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals
import unittest

from hamcrest import assert_that, equal_to, none, only_contains, raises

from storops_test.vnx.nas_mock import t_nas, patch_post, patch_nas
from storops.vnx.enums import VNXPortType
from storops.exception import VNXBackendError, VNXGeneralNasError
from storops.vnx.resource.mover import VNXMover, \
    VNXMoverRefList, VNXMoverRef, VNXMoverHost, VNXMoverHostList

__author__ = 'Jay Xu'


class VNXMoverRefTest(unittest.TestCase):
    @patch_post
    def test_get_all(self):
        movers = VNXMoverRefList(t_nas())
        assert_that(len(movers), equal_to(2))
        dm = next(dm for dm in movers if dm.mover_id == 1)
        self.verify_dm_ref_1(dm)

    @patch_post
    def test_get(self):
        dm = VNXMoverRef(mover_id=1, cli=t_nas())
        self.verify_dm_ref_1(dm)

    @patch_post
    def test_get_not_existed(self):
        dm = VNXMoverRef(mover_id=5, cli=t_nas())
        assert_that(dm.existed, equal_to(False))

    @patch_post
    def test_get_by_name(self):
        dm = VNXMoverRef(name='server_2', cli=t_nas())
        self.verify_dm_ref_1(dm)

    @patch_post
    def test_get_by_name_not_found(self):
        dm = VNXMoverRef(name='server_5', cli=t_nas())
        assert_that(dm.existed, equal_to(False))

    @staticmethod
    def verify_dm_ref_1(dm):
        assert_that(dm.mover_id, equal_to(1))
        assert_that(dm.i18n_mode, equal_to('UNICODE'))
        assert_that(dm.name, equal_to('server_2'))
        assert_that(dm.existed, equal_to(True))
        assert_that(dm.standby_fors, none())
        assert_that(dm.failover_policy, equal_to('auto'))
        assert_that(dm.host_id, equal_to(1))
        assert_that(dm.role, equal_to('primary'))
        assert_that(dm.standbys, only_contains(2))

    def test_get_id(self):
        dm = VNXMover(mover_id=12)
        assert_that(dm.get_id(dm), equal_to(12))
        assert_that(dm.get_id('22'), equal_to(22))

    @patch_post
    def test_mover_host(self):
        dm = VNXMover(mover_id=1, cli=t_nas())
        VNXMoverHostTest.verify_mover_host_1(dm.host)

    @patch_post
    def test_create_dns(self):
        dm = VNXMover.get(mover_id=1, cli=t_nas())
        resp = dm.create_dns('tt', '1.1.1.1')
        assert_that(resp.is_ok(), equal_to(True))

    @patch_post
    def test_create_dns_format_error(self):
        def f():
            dm = VNXMover(mover_id=1, cli=t_nas())
            dm.create_dns('tt', '1.1.1.1,2.2.2.2')

        assert_that(f, raises(VNXBackendError, 'not facet-valid'))

    @patch_post
    def test_create_dns_multiple(self):
        dm = VNXMover(mover_id=1, cli=t_nas())
        resp = dm.create_dns('tt', ['1.1.1.1', '2.2.2.2'])
        assert_that(resp.is_ok(), equal_to(True))

    @patch_post
    def test_delete_dns(self):
        dm = VNXMoverRef(mover_id=1, cli=t_nas())
        resp = dm.delete_dns('tt')
        assert_that(resp.is_ok(), equal_to(True))

    @patch_post
    def test_delete_dns_not_exist(self):
        def f():
            dm = VNXMoverRef(mover_id=1, cli=t_nas())
            dm.delete_dns('bb')

        assert_that(f, raises(VNXGeneralNasError, 'server_2'))

    @patch_post
    def test_physical_devices(self):
        dm = VNXMoverRef(mover_id=1, cli=t_nas())
        assert_that(len(dm.physical_devices), equal_to(9))

    @patch_post
    def test_fc_devices(self):
        dm = VNXMoverRef(mover_id=1, cli=t_nas())
        assert_that(len(dm.fc_devices), equal_to(4))

    @patch_post
    def test_ethernet_devices(self):
        dm = VNXMoverRef(mover_id=1, cli=t_nas())
        assert_that(len(dm.ethernet_devices), equal_to(4))

    @patch_post
    def test_create_interface(self):
        dm = VNXMover(mover_id=1, cli=t_nas())
        interface = dm.create_interface('cge-1-0', '1.1.1.1', '255.255.255.0')
        assert_that(interface.name, equal_to('1.1.1.1-0'))
        assert_that(interface.broadcast_addr, equal_to('1.1.1.255'))

    @patch_post
    def test_delete_interface(self):
        dm = VNXMover(mover_id=1, cli=t_nas())
        resp = dm.delete_interface('1.1.1.1')
        assert_that(resp.is_ok(), equal_to(True))

    @patch_nas
    def test_get_interconnect_id(self):
        dm = VNXMover(mover_id=1, cli=t_nas())
        assert_that(dm.get_interconnect_id(), equal_to(20001))


class VNXMoverTest(unittest.TestCase):
    @patch_post
    def test_create_nfs_share(self):
        dm = VNXMover(mover_id=1, cli=t_nas())
        share = dm.create_nfs_share(path='/EEE')
        assert_that(share.path, equal_to('/EEE'))
        assert_that(share.mover_id, equal_to(1))
        assert_that(share.existed, equal_to(True))
        assert_that(share.fs_id, equal_to(243))

    @patch_post
    def test_get_all(self):
        movers = VNXMover.get(t_nas())
        assert_that(len(movers), equal_to(2))
        mover1 = next(dm for dm in movers if dm.mover_id == 1)
        self.verify_dm_1(mover1)

    @patch_post
    def test_get(self):
        mover1 = VNXMover.get(mover_id=1, cli=t_nas())
        self.verify_dm_1(mover1)

    @staticmethod
    def verify_dm_1(dm):
        assert_that(dm.status, equal_to('ok'))
        assert_that(dm.mover_id, equal_to(1))
        assert_that(dm.uptime, equal_to(7086723))
        assert_that(dm.i18n_mode, equal_to('UNICODE'))
        assert_that(dm.name, equal_to('server_2'))
        assert_that(dm.existed, equal_to(True))
        assert_that(dm.version, equal_to('T8.1.7.70'))
        assert_that(dm.standby_fors, none())
        assert_that(dm.dns_domain, none())
        assert_that(dm.failover_policy, equal_to('auto'))
        assert_that(dm.host_id, equal_to(1))
        assert_that(dm.role, equal_to('primary'))
        assert_that(dm.standbys, only_contains(2))
        assert_that(dm.timezone, equal_to('GMT'))
        interface = next(i for i in dm.interfaces if i.name == 'el30')
        assert_that(interface.mover_id, equal_to(1))
        assert_that(interface.ip_addr, equal_to('172.18.70.2'))
        assert_that(interface.name, equal_to('el30'))
        assert_that(interface.existed, equal_to(True))
        assert_that(interface.broadcast_addr, equal_to('172.18.255.255'))
        assert_that(interface.net_mask, equal_to('255.255.0.0'))
        assert_that(interface.up, equal_to(True))
        assert_that(interface.mtu, equal_to(1500))
        assert_that(interface.ip_version, equal_to('IPv4'))
        assert_that(interface.mac_addr, equal_to('2:60:48:20:b:0'))
        assert_that(interface.device, equal_to('cge0'))
        assert_that(interface.vlan_id, equal_to(0))
        dedup_settings = dm.dedup_settings
        assert_that(dedup_settings.cpu_high_watermark, equal_to(90))
        assert_that(dedup_settings.minimum_scan_interval, equal_to(7))
        assert_that(dedup_settings.duplicate_detection_method,
                    equal_to('sha1'))
        assert_that(dedup_settings.mover_id, equal_to(1))
        assert_that(dedup_settings.minimum_size, equal_to(24))
        assert_that(dedup_settings.access_time, equal_to(15))
        assert_that(dedup_settings.file_extension_exclude_list, equal_to(''))
        assert_that(dedup_settings.case_sensitive, equal_to(False))
        assert_that(dedup_settings.cifs_compression_enabled, equal_to(True))
        assert_that(dedup_settings.modification_time, equal_to(15))
        assert_that(dedup_settings.sav_vol_high_watermark, equal_to(90))
        assert_that(dedup_settings.backup_data_high_watermark, equal_to(90))
        assert_that(dedup_settings.maximum_size, equal_to(8388608))
        assert_that(dedup_settings.cpu_low_watermark, equal_to(40))
        assert_that(dedup_settings.existed, equal_to(True))
        device = next(i for i in dm.devices if i.name == 'fxg-8-0')
        assert_that(device.mover_id, equal_to(1))
        assert_that(device.name, equal_to('fxg-8-0'))
        assert_that(device.existed, equal_to(True))
        assert_that(device.type, equal_to('physical-ethernet'))
        assert_that(device.interfaces, equal_to('10.110.42.83'))
        assert_that(device.speed, equal_to('FD10000'))
        route = next(i for i in dm.route if i.destination == '172.18.0.0')
        assert_that(route.mover_id, equal_to(1))
        assert_that(route.existed, equal_to(True))
        assert_that(route.destination, equal_to('172.18.0.0'))
        assert_that(route.net_mask, equal_to('255.255.0.0'))
        assert_that(route.ip_version, equal_to('IPv4'))
        assert_that(route.interface, equal_to('172.18.70.2'))
        assert_that(route.gateway, equal_to('172.18.70.2'))


class VNXMoverHostTest(unittest.TestCase):
    @patch_post
    def test_get_by_id(self):
        mh = VNXMoverHost(host_id=1, cli=t_nas())
        self.verify_mover_host_1(mh)

    @patch_post
    def test_get_all(self):
        mh_list = VNXMoverHostList(t_nas())
        assert_that(len(mh_list), equal_to(2))
        mh = next(mh for mh in mh_list if mh.host_id == 1)
        self.verify_mover_host_1(mh)

    @patch_post
    def test_get_by_id_not_found(self):
        mh = VNXMoverHost(host_id=5, cli=t_nas())
        assert_that(mh.existed, equal_to(False))

    @classmethod
    def verify_mover_host_1(cls, mh):
        assert_that(mh.host_id, equal_to(1))
        assert_that(mh.slot, equal_to(2))
        assert_that(mh.existed, equal_to(True))
        assert_that(mh.mover_id, equal_to(1))
        assert_that(mh.status, equal_to('ok'))
        motherboard = mh.motherboard
        assert_that(motherboard.existed, equal_to(True))
        assert_that(motherboard.mover_host, equal_to(1))
        assert_that(motherboard.bus_speed, equal_to(4800))
        assert_that(motherboard.memory_size, equal_to(12288))
        assert_that(motherboard.cpu_type, equal_to('Intel Four Core Westmere'))
        assert_that(motherboard.cpu_speed, equal_to(2133))
        assert_that(motherboard.board_type, equal_to('CMB-Argonaut'))
        fcp02 = next(device for device in mh.physical_device
                     if device.name == 'fcp-0-2')
        cls.verify_fcp_02(fcp02)
        cge13 = next(device for device in mh.physical_device
                     if device.name == 'cge-1-3')
        cls.verify_cge13(cge13)

    @staticmethod
    def verify_cge13(cge13):
        assert_that(cge13.existed, equal_to(True))
        assert_that(cge13.name, equal_to('cge-1-3'))
        assert_that(cge13.irq, equal_to(27))
        assert_that(cge13.mover_host, equal_to(1))
        assert_that(cge13.port_number, equal_to(0))
        assert_that(cge13.allowed_speeds,
                    equal_to('FD1000 FD100 HD100 FD10 HD10 auto'))
        assert_that(cge13.port_wwn, none())
        assert_that(cge13.type, equal_to(VNXPortType.ETHERNET))
        assert_that(cge13.description, equal_to('Broadcom Gigabit'))
        assert_that(cge13.is_internal, equal_to(False))

    @staticmethod
    def verify_fcp_02(fcp02):
        assert_that(fcp02.existed, equal_to(True))
        assert_that(fcp02.name, equal_to('fcp-0-2'))
        assert_that(fcp02.irq, equal_to(22))
        assert_that(fcp02.mover_host, equal_to(1))
        assert_that(fcp02.port_number, equal_to(0))
        assert_that(fcp02.allowed_speeds, none())
        assert_that(fcp02.port_wwn, equal_to('50:06:01:62:47:60:44:06'))
        assert_that(fcp02.type, equal_to(VNXPortType.FC))
        assert_that(fcp02.description, equal_to('PMC QE8'))
        assert_that(fcp02.is_internal, equal_to(False))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

from hamcrest import assert_that, instance_of, has_items, equal_to, raises, \
    is_not

from storops.lib.common import instance_cache
from storops.lib.resource import ResourceListCollection
from storops.unity.resource.lun import UnityLun
from storops.unity.resource.sp import UnityStorageProcessor, \
    UnityStorageProcessorList
from storops_test.unity.rest_mock import t_unity, patch_rest

__author__ = 'Cedric Zhuang'


class ResourceListCollectionTest(unittest.TestCase):
    @instance_cache
    def get_rlc(self):
        return ResourceListCollection(
            (t_unity().get_sp(), t_unity().get_lun()))

    def test_get_rsc_list(self):
        rsc_list = self.get_rlc().get_rsc_list(UnityStorageProcessor)
        assert_that(rsc_list, instance_of(UnityStorageProcessorList))

    def test_get_rsc_clz_list(self):
        clz_list = self.get_rlc().get_rsc_clz_list()
        assert_that(clz_list, has_items(UnityStorageProcessor, UnityLun))

    def test_len(self):
        assert_that(len(self.get_rlc()), equal_to(2))

    def test_add_rsc_list_wrong_type(self):
        def f():
            self.get_rlc().add_rsc_list('abc')

        assert_that(f, raises(ValueError))

    def test_add_rsc_list_success(self):
        rlc = self.get_rlc()
        rlc.add_rsc_list(t_unity().get_disk())
        assert_that(len(rlc), equal_to(3))

    @patch_rest
    def test_update(self):
        rlc = self.get_rlc()
        t0 = rlc.timestamp
        rlc.update()
        assert_that(t0, is_not(equal_to(rlc.timestamp)))


class ResourceListTest(unittest.TestCase):
    @patch_rest
    def test_add_resource_list(self):
        ret = t_unity().get_ethernet_port() + t_unity().get_link_aggregation()
        assert_that(len(ret), equal_to(10))

    def test_add_resource_list_type_error(self):
        def do():
            return t_unity().get_ethernet_port() + None
        assert_that(do, raises(TypeError))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import pytest

from storops_comptest.unity import UnityGeneralFixture, UnityCifsShareFixture
from storops_comptest.utils import is_jenkins, setup_fixture

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


@pytest.fixture(scope="session", autouse=True)
def unity_gf(request):
    """ General fixture for most unity cases

    :param request:
    :return:
    """
    return setup_fixture(request, UnityGeneralFixture)


@pytest.fixture(scope="session", autouse=True)
def unity_cs(request):
    """ General fixture for unity cifs share in domain

    :param request:
    :return:
    """
    if is_jenkins():
        pytest.skip('do not run on CI, manual only.')

    return setup_fixture(request, UnityCifsShareFixture)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

from hamcrest import assert_that, equal_to, has_items

from storops_test.vnx.nas_mock import t_nas, patch_post
from storops.vnx.resource.nas_pool import VNXNasPool

__author__ = 'Jay Xu'


class VNXNasPoolTest(unittest.TestCase):
    @patch_post
    def test_get_all(self):
        pool_list = VNXNasPool.get(cli=t_nas())
        assert_that(len(pool_list), equal_to(6))

    @patch_post
    def test_get_by_name(self):
        pool = VNXNasPool.get(name='vnx-sg_test', cli=t_nas())
        self.verify_pool_vnx_sg_test(pool)

    @patch_post
    def test_get_by_pool_id(self):
        pool = VNXNasPool.get(pool_id=63, cli=t_nas())
        self.verify_pool_vnx_sg_test(pool)

    @patch_post
    def test_get_by_name_not_found(self):
        pool = VNXNasPool(name='not_found', cli=t_nas())
        assert_that(pool.existed, equal_to(False))

    def verify_pool_vnx_sg_test(self, pool):
        assert_that(pool.movers_id, has_items(1, 2))
        assert_that(pool.member_volumes, has_items(105))
        assert_that(pool.name, equal_to('vnx-sg_test'))
        assert_that(pool.description, equal_to("vnx-sg_test on 000196800192"))
        assert_that(pool.may_contain_slices_default, equal_to(False))
        assert_that(pool.disk_type, equal_to('Mixed'))
        assert_that(pool.size, equal_to(0))
        assert_that(pool.used_size, equal_to(0))
        assert_that(pool.total_size, equal_to(2077))
        assert_that(pool.virtual_provisioning, equal_to(True))
        assert_that(pool.is_homogeneous, equal_to(True))
        assert_that(pool.template_pool, equal_to(63))
        assert_that(pool.stripe_count, equal_to(8))
        assert_that(pool.stripe_size, equal_to(256))
        assert_that(pool.pool_id, equal_to(63))

    def test_get_id(self):
        pool = VNXNasPool(pool_id=12)
        assert_that(pool.get_id(pool), equal_to(12))
        assert_that(pool.get_id('22'), equal_to(22))



# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Veriion 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import unittest

import hamcrest as hc

import storops
from storops.unity.resource import remote_interface, remote_system
from storops_test.unity import rest_mock as mock

__author__ = 'Ryan Liang'


class UnityRemoteInterfaceTest(unittest.TestCase):
    @mock.patch_rest
    def test_get_properties(self):
        ri = remote_interface.UnityRemoteInterface.get(
            cli=mock.t_rest(), _id='FNM00150600267:if_10')
        hc.assert_that(ri.address, hc.equal_to('10.245.47.99'))
        hc.assert_that(ri.capability,
                       hc.equal_to(storops.ReplicationCapabilityEnum.ASYNC))
        hc.assert_that(ri.id, hc.equal_to('FNM00150600267:if_10'))
        hc.assert_that(ri.name, hc.equal_to('4_FNM00150600267'))
        hc.assert_that(ri.node, hc.equal_to(storops.NodeEnum.SPA))
        hc.assert_that(ri.remote_id, hc.equal_to('if_10'))
        hc.assert_that(ri.remote_system,
                       hc.instance_of(remote_system.UnityRemoteSystem))
        hc.assert_that(ri.remote_system.id, hc.equal_to('RS_0'))

    @mock.patch_rest
    def test_get_collection(self):
        ri_list = remote_interface.UnityRemoteInterfaceList.get(mock.t_rest())
        hc.assert_that(len(ri_list), hc.equal_to(8))
        hc.assert_that([ri.id for ri in ri_list],
                       hc.equal_to(['FNM00150600267:if_10',
                                    'FNM00150600267:if_11',
                                    'FNM00150600267:system_if_0',
                                    'FNM00150600267:system_if_1',
                                    'FNM00150600267:system_if_2',
                                    'FNM00150600267:system_if_3',
                                    'FNM00152000052:if_6',
                                    'FNM00152000052:if_7']))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, raises, contains_string, \
    only_contains

from storops.exception import VNXNasObjectNotFound, \
    VNXInvalidMoverID, VNXFileCredentialError
from storops.vnx.nas_client import NasXmlResponse, XmlStatus
from storops_test.vnx.nas_mock import MockXmlPost

__author__ = 'Cedric Zhuang'


class NasXmlResponseTest(TestCase):
    @property
    def fs_not_found(self):
        return NasXmlResponse(MockXmlPost.read_file('fs_not_found.xml'))

    @property
    def invalid_dm(self):
        return NasXmlResponse(MockXmlPost.read_file('invalid_data_mover.xml'))

    def test_is_ok(self):
        resp = self.fs_not_found
        assert_that(resp.is_ok(), equal_to(False))

    def test_credential_error(self):
        def f():
            NasXmlResponse(MockXmlPost.read_file('credential_error.html'))

        assert_that(f, raises(VNXFileCredentialError, 'credential error'))

    def test_has_error_code(self):
        resp = self.fs_not_found
        assert_that(resp.has_error_code('18522112101'), equal_to(True))
        assert_that(resp.has_error_code(18522112101), equal_to(True))
        assert_that(resp.has_error_code(18522112102), equal_to(False))

    def test_status(self):
        resp = self.fs_not_found
        assert_that(resp.status, equal_to(XmlStatus.WARNING))

    def test_problem_message_codes(self):
        resp = self.fs_not_found
        assert_that(resp.problem_message_codes, equal_to({18522112101, }))

    def test_problem_diagnostics(self):
        resp = self.fs_not_found
        assert_that(resp.problem_diagnostics, contains_string(
            'Migration file system not found.'))
        assert_that(resp.problem_diagnostics, contains_string(
            'File system not found.'))

    def test_problem_messages(self):
        resp = self.fs_not_found
        assert_that(resp.problem_messages, equal_to(
            'The query may be incomplete or requested object not found.'))

    def test_check_invalid_data_mover(self):
        def f():
            resp = self.invalid_dm
            resp.check_invalid_data_mover()

        assert_that(f, raises(VNXInvalidMoverID, 'id=100 not found'))

    def test_raise_if_not_found(self):
        def f():
            resp = self.fs_not_found
            resp.raise_if_no_object('no fs here')

        assert_that(f, raises(VNXNasObjectNotFound, 'no fs here'))

    def test_get_problem_string(self):
        resp = self.fs_not_found
        problems = resp.get_problems_string(' ' * 4)
        assert_that(problems, contains_string('    (1) code: 18522112101'))
        assert_that(problems, contains_string('        message: The query'))
        assert_that(problems, contains_string('        diagnostics:'))

    def test_problem_message_codes_hex(self):
        resp = self.fs_not_found
        assert_that(resp.hex_problem_message_codes,
                    only_contains('0x450010065'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import six

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class MappedErrorCodeDecoratorFactory(object):
    def __init__(self, default_exception, exception_map=None):
        if exception_map is None:
            exception_map = {}
        self.exception_map = exception_map
        self.default = default_exception

    def clz_decorator(self):
        def decorator(clz):
            if not hasattr(clz, 'error_code'):
                raise AttributeError(
                    '"error_code" property is required on class {}.'.format(
                        clz.__name__))
            if not isinstance(clz.error_code, (list, tuple, set)):
                error_codes = [clz.error_code]
            else:
                error_codes = clz.error_code

            for error_code in error_codes:
                if error_code not in self.exception_map:
                    self.exception_map[error_code] = clz
            return clz

        return decorator

    def get_exception(self, error_code, default=None):
        if default is None:
            default = self.default
        ret = default
        if error_code is not None:
            if not isinstance(error_code, (list, tuple, set)):
                error_code = [error_code]
            for code in error_code:
                if code in self.exception_map:
                    ret = self.exception_map[code]
                    break
        return ret


class ExceptionListDecoratorFactory(object):
    def __init__(self, default_exception, exception_list=None):
        if exception_list is None:
            exception_list = []
        self.exception_list = exception_list
        self.default = default_exception

    def clz_decorator(self):
        def decorator(clz):
            self.exception_list.append(clz)
            return clz

        return decorator

    def get_exception(self, output, default=None):
        if default is not None:
            ret = default
        else:
            ret = self.default

        if output:
            for clz in self.exception_list:
                msg = clz.get_error_message()
                if isinstance(msg, (tuple, list, set)):
                    found = any(m in output for m in msg)
                elif isinstance(msg, six.string_types):
                    found = msg in output
                elif hasattr(msg, 'search'):
                    found = msg.search(output) is not None
                else:
                    raise ValueError('{} is not a valid message.'.format(msg))

                if found:
                    ret = clz
                    break
        return ret



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.unity.resource import UnityResource, UnityResourceList
from storops.unity.resource.storage_resource import UnityStorageResource

__author__ = 'Cedric Zhuang'


class UnityDataStore(UnityResource):
    @classmethod
    def create(cls, cli, name, vvol_datastore_type, description=None,
               hosts=None,
               capability_profiles=None):
        vvol_ds_cp_parameters = {"addCapabilityProfile": [
            {"capProfile": {"id": x.id}, "sizeTotal": x.pool.size_total} for x
            in capability_profiles]}

        req_body = cli.make_body(
            name=name,
            vvolDatastoreType=vvol_datastore_type,
            description=description,
            hosts=hosts,
            vvolDatastoreCapabilityProfilesParameters=vvol_ds_cp_parameters)

        resp = cli.type_action(UnityStorageResource().resource_class,
                               'createVVolDatastore', **req_body)
        resp.raise_if_err()
        return UnityStorageResource(_id=resp.resource_id, cli=cli)


class UnityDataStoreList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityDataStore


class UnityVmDisk(UnityResource):
    pass


class UnityVmDiskList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityVmDisk


class UnityVm(UnityResource):
    pass


class UnityVmList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityVm


class UnityVirtualVolume(UnityResource):
    pass


class UnityVirtualVolumeList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityVirtualVolume


class UnityHostVvolDatastore(UnityResource):
    pass


class UnityHostVvolDatastoreList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityHostVvolDatastore


class UnityCapabilityProfile(UnityResource):
    @classmethod
    def create(cls, cli, name, pool, description=None,
               usage_tags=None, constraints=None):
        req_body = cli.make_body(name=name,
                                 pool=pool,
                                 description=description,
                                 usageTags=usage_tags,
                                 constraints=constraints)
        resp = cli.post(cls().resource_class, **req_body)
        resp.raise_if_err()
        return cls.get(cli, resp.resource_id)


class UnityCapabilityProfileList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityCapabilityProfile


class UnityVirtualVolumeBinding(UnityResource):
    pass


class UnityVirtualVolumeBindingList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityVirtualVolume


class UnityVmwarePE(UnityResource):
    pass


class UnityVmwarePEList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityVmwarePE


class UnityVmwareNasPEServer(UnityResource):
    pass


class UnityVmwareNasPEServerList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityVmwareNasPEServer



# coding=utf-8
# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.vnx.resource.disk import VNXDisk
from storops.vnx.resource.block_pool import VNXPool
from storops.vnx.resource.rg import VNXRaidGroup
from storops.lib.common import JsonPrinter, round_3
from storops.lib.converter import mb_to_gb

__author__ = 'Tina Tang'


class VNXCapacity(JsonPrinter):

    _properties = ['used', 'total', 'free_raw_disk', 'free_storage_pool']

    def __init__(self, cli=None):
        super(VNXCapacity, self).__init__()
        self._cli = cli
        self._total = None
        self._free_raw_disk = None
        self._free_storage_pool = None
        self._used = None

    @classmethod
    def get(cls, cli):
        return VNXCapacity(cli)

    @property
    @round_3
    def total(self):
        """Total disk capacity in GB"""
        if self._total is None:
            self._update_disk_capacity()
        return self._total

    @property
    @round_3
    def free_raw_disk(self):
        """Free raw disk capacity in GB

        Physical capacity (in GB) of all unused disks, including disks in empty
        RAID Groups.
        """
        if self._free_raw_disk is None:
            self._update_disk_capacity()
        return self._free_raw_disk

    @property
    @round_3
    def free_storage_pool(self):
        """Free storage pool capacity in GB

        Free storage capacity (in GB) in all storage pools and non-empty RAID
        Groups.
        """
        if self._free_storage_pool is None:
            self._update_free_pool_capacity()
        return self._free_storage_pool

    @property
    @round_3
    def used(self):
        """Total used capacity in GB

        Capacity consumed by LUNs, RAID protection and storage pools for File.
        """
        return self.total - self.free_storage_pool - self.free_raw_disk

    def update(self):
        self._update_free_pool_capacity()
        self._update_disk_capacity()

    def _update_disk_capacity(self):
        disks = VNXDisk.get(self._cli)
        self._total = mb_to_gb(sum([cap for cap in disks.capacity if cap is
                                    not None]))
        self._free_raw_disk = mb_to_gb(
            sum([disk.capacity for disk in disks if disk.state == 'Unbound']))

    def _update_free_pool_capacity(self):
        pools = VNXPool.get(self._cli)
        free_in_pool = sum(pools.available_capacity_gbs)
        rgs = VNXRaidGroup.get(self._cli)
        free_in_rg = sum([rg.available_capacity_gbs for rg in rgs
                          if len(rg.list_of_luns) != 0])
        self._free_storage_pool = free_in_pool + free_in_rg

    def _get_properties(self, dec=0):
        re = {}
        for props in self._properties:
            re[props] = getattr(self, props)
        return re



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from hamcrest import assert_that, equal_to, has_item

__author__ = 'Cedric Zhuang'


def test_sg_create_success(vnx_gf):
    assert_that(vnx_gf.sg.existed, equal_to(True))


def test_sg_attach_detach_lun(vnx_gf):
    sg = vnx_gf.sg
    lun = vnx_gf.lun

    sg.attach_alu(lun)
    # check cache
    assert_that(sg.get_hlu(lun), equal_to(1))

    sg.update()
    # check cli parsing
    assert_that(sg.get_hlu(lun), equal_to(1))

    sg.detach_alu(lun)
    # check cache
    assert_that(sg.has_alu(lun), equal_to(False))

    sg.update()
    # check cli parsing
    assert_that(sg.has_alu(lun), equal_to(False))


def test_set_delete_path(vnx_gf):
    sp, port_id = vnx_gf.select_port()
    uid = 'iqn.1992-04.com.abc:a.b.c'

    port = vnx_gf.vnx.get_iscsi_port(sp, port_id)[0]
    port = port.config_ip('7.7.7.7', '255.255.255.0', '7.7.7.1', 3, 3)

    vnx_gf.until_existed(port)

    sg = vnx_gf.vnx.create_sg(vnx_gf.add_sg_name())
    sg.set_path(port, uid, 'host0')

    sg.update()
    assert_that(sg.hba_sp_pairs.host_name, has_item('host0'))

    sg.delete(disconnect_host=True)
    vnx_gf.vnx.delete_hba(uid)
    port.delete_ip()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, has_item, raises, only_contains

from storops.vnx.parsers import get_vnx_parser
from storops_test.vnx.cli_mock import patch_cli, t_cli
from storops.exception import VNXConsistencyGroupError, \
    VNXConsistencyGroupNameInUseError, VNXConsistencyGroupNotFoundError, \
    VNXSnapNameInUseError
from storops.vnx.resource.cg import VNXConsistencyGroup
from storops.vnx.resource.cg import VNXConsistencyGroupList
from storops.vnx.resource.lun import VNXLun

__author__ = 'Cedric Zhuang'


class VNXConsistencyGroupListTest(TestCase):
    @patch_cli
    def test_parse(self):
        assert_that(len(VNXConsistencyGroupList(t_cli())), equal_to(2))

    @patch_cli
    def test_delete_not_existed_member(self):
        lun = VNXLun(name='y', cli=t_cli())
        cg_list = VNXConsistencyGroupList(t_cli())
        # no error raised
        cg_list.delete_member(lun)


class VNXConsistencyGroupTest(TestCase):
    @patch_cli
    def test_list_consistency_group(self):
        cg_list = VNXConsistencyGroup.get(t_cli())
        assert_that(len(cg_list), equal_to(2))
        assert_that(cg_list.name, only_contains('another cg', 'test cg name'))

    @patch_cli
    def test_update_consistency_group_list_member(self):
        cg_list = VNXConsistencyGroup.get(t_cli())
        cg = cg_list[1]
        name = cg.name
        cg.update()
        assert_that(cg.name, equal_to(name))

    @patch_cli
    def test_properties(self):
        cg = VNXConsistencyGroup(name="test_cg", cli=t_cli())
        assert_that(cg.name, equal_to('test_cg'))
        assert_that(cg.lun_list.lun_id, only_contains(1, 3))
        assert_that(cg.state, equal_to('Ready'))
        assert_that(cg.existed, equal_to(True))

    def test_update(self):
        parser = get_vnx_parser('VNXConsistencyGroup')
        data = {
            parser.NAME.key: 'test cg name',
            parser.LUN_LIST.key: [1, 5, 7],
            parser.STATE.key: 'Offline'
        }

        cg = VNXConsistencyGroup()
        cg.update(data)

        assert_that(cg.name, equal_to('test cg name'))
        assert_that(cg.lun_list, only_contains(1, 5, 7))
        assert_that(cg.state, equal_to('Offline'))

    def test_parse(self):
        output = """
                Name:  test cg name
                Name:  another cg
                """
        cgs = VNXConsistencyGroup.parse_all(output)
        assert_that(len(cgs), equal_to(2))
        names = [cg.name for cg in cgs]
        assert_that(names, has_item('test cg name'))
        assert_that(names, has_item('another cg'))

    @patch_cli
    def test_add_member(self):
        def f():
            cg = VNXConsistencyGroup('test_cg', t_cli())
            m1 = VNXLun(name='m1', cli=t_cli())
            m2 = VNXLun(name='m2', cli=t_cli())
            cg.add_member(m1, m2)

        assert_that(f, raises(VNXConsistencyGroupError, 'Cannot add members'))

    @patch_cli
    def test_has_member(self):
        cg = VNXConsistencyGroup('test_cg', t_cli())
        lun = VNXLun(lun_id=1)
        assert_that(cg.has_member(lun), equal_to(True))
        assert_that(cg.has_member(7), equal_to(False))

    @patch_cli
    def test_cg_no_poll(self):
        def f():
            cg = VNXConsistencyGroup(name="test_cg", cli=t_cli())
            with cg.with_no_poll():
                cg.add_member(1, 2, 3)

        assert_that(f, raises(VNXConsistencyGroupError, 'does not exist'))

    @patch_cli
    def test_cg_not_found(self):
        def f():
            cg = VNXConsistencyGroup(name="cg1", cli=t_cli())
            cg.add_member(1)

        assert_that(f, raises(VNXConsistencyGroupNotFoundError, 'Cannot find'))

    @patch_cli
    def test_create_cg_name_in_use(self):
        def f():
            VNXConsistencyGroup.create(cli=t_cli(), name='cg0')

        assert_that(f, raises(VNXConsistencyGroupNameInUseError,
                              'already in use'))

    @patch_cli
    def test_delete_cg_not_exists(self):
        def f():
            cg = VNXConsistencyGroup(cli=t_cli(), name='cg0')
            cg.delete()

        assert_that(f, raises(VNXConsistencyGroupNotFoundError, 'Cannot find'))

    @patch_cli
    def test_create_cg_snap_success(self):
        cg = VNXConsistencyGroup(name="cg1", cli=t_cli())
        snap = cg.create_snap('cg1_snap')
        assert_that(snap._name, equal_to('cg1_snap'))
        assert_that(snap.source_luns, only_contains(1))
        assert_that(snap.source_cg, equal_to('cg1'))

    @patch_cli
    def test_create_cg_snap_name_existed(self):
        def f():
            cg = VNXConsistencyGroup(name="cg2", cli=t_cli())
            cg.create_snap('cg1_snap')

        assert_that(f, raises(VNXSnapNameInUseError, 'already in use'))

    @patch_cli
    def test_cg_empty_member_property(self):
        cg = VNXConsistencyGroup(name='cg0', cli=t_cli())
        assert_that(len(cg.lun_list), equal_to(0))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import equal_to, assert_that, none, has_items, only_contains

from storops.vnx.resource.host import VNXHostList, VNXHost
from storops_test.vnx.cli_mock import t_cli, patch_cli

__author__ = 'Cedric Zhuang'


class VNXHostTest(TestCase):
    @patch_cli
    def test_get_all(self):
        host_list = VNXHostList(t_cli())
        assert_that(len(host_list), equal_to(6))
        assert_that(host_list.name, has_items(
            'APM00152312055-spB', 'APM00152312055-spA', 'ubuntu-server11',
            'ubuntu-server7', 'Celerra_CS0_21132', 'ubuntu14'))

    @patch_cli
    def test_name_filter_of_get_all(self):
        host_list = VNXHostList(t_cli(), names=('ubuntu-server11', 'ubuntu14'))
        assert_that(len(host_list), equal_to(2))
        assert_that(host_list.name, has_items('ubuntu-server11', 'ubuntu14'))

    @patch_cli
    def test_host_property(self):
        host = VNXHost.get(cli=t_cli(), name='ubuntu-server7')
        assert_that(host.name, equal_to('ubuntu-server7'))
        assert_that(host.existed, equal_to(True))
        assert_that(len(host.connections), equal_to(15))
        assert_that(host.storage_group.name, equal_to('ubuntu-server7'))
        assert_that(len(host.lun_list), equal_to(0))

    @patch_cli
    def test_host_with_lun(self):
        host = VNXHost.get(cli=t_cli(), name='ubuntu14')
        assert_that(host.lun_list.lun_id, only_contains(4, 15))
        assert_that(host.alu_hlu_map[4], equal_to(14))
        assert_that(host.alu_hlu_map[15], equal_to(154))
        assert_that(host.alu_ids, only_contains(4, 15))
        assert_that(host.hlu_ids, only_contains(14, 154))

    @patch_cli
    def test_host_not_found(self):
        host = VNXHost.get(cli=t_cli(), name='not_found')
        assert_that(host.existed, equal_to(False))
        assert_that(host.name, equal_to('not_found'))
        assert_that(host.connections, none())



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import functools
import logging
import re

import six
from retryz import retry

from storops.connection import connector
from storops.exception import VNXBackendError, VNXLockRequiredException, \
    VNXNasObjectNotFound, VNXInvalidMoverID, VNXException, \
    get_xmlapi_exception, VNXFileCredentialError
from storops.lib.common import Enum, check_int
from storops.lib.converter import to_int, to_hex
from storops.vnx.nas_cmd import NasCommand
from storops.vnx.resource.cifs_share import CifsAccessControl
from storops.vnx.xmlapi import NasXmlBuilder
from storops.vnx.xmlapi_parser import XMLAPIParser

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


class VNXNasConnections(object):
    retry_patterns = [(r'unable to acquire lock\(s\)',
                       VNXLockRequiredException())]

    def __init__(self, host, username, password, ssh_port=22):
        self.host = host
        self.username = username
        self.password = password
        self.ssh_port = ssh_port

        self._ssh_timeout = None

        self._ssh = None
        self._xml_connector = None
        self._xml_parser = None

    @property
    def ssh(self):
        if self._ssh is None:
            self._ssh = self._ssh = connector.SSHConnector(
                self.host, self.username, self.password, self.ssh_port)
        return self._ssh

    def set_ssh_timeout(self, value):
        self._ssh_timeout = check_int(value)

    @property
    def xml_connector(self):
        if self._xml_connector is None:
            self._xml_connector = connector.XMLAPIConnector(
                self.host, self.username, self.password)
        return self._xml_connector

    @property
    def xml_parser(self):
        if self._xml_parser is None:
            self._xml_parser = XMLAPIParser()
        return self._xml_parser

    @classmethod
    def _get_req_xml(cls, req):
        base = '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>{}'
        return base.format(req)

    @retry(on_error=VNXLockRequiredException)
    def _request(self, req, retry_patterns=None):

        req_xml = self._get_req_xml(req)
        log.debug('request: \n{}'.format(req_xml))

        rsp_xml = self.xml_connector.post(req_xml)

        if isinstance(rsp_xml, tuple):
            rsp_xml = rsp_xml[1]
        log.debug('response: \n{}'.format(rsp_xml))

        response = NasXmlResponse(self.xml_parser.parse(rsp_xml))

        if not retry_patterns:
            retry_patterns = self.retry_patterns
        if response.is_error():
            for pattern in retry_patterns:
                messages = response.problem_messages
                to_match, to_raise = pattern
                if re.search(to_match, messages):
                    raise to_raise

        return response

    def request(self, req, check_object=False, check_invalid_data_mover=False,
                error_desc=None, retry_patterns=None):
        response = self._request(req, retry_patterns)

        if check_object:
            response.raise_if_no_object(error_desc)

        if check_invalid_data_mover:
            response.check_invalid_data_mover()

        return response

    def ssh_execute(self, commands, check_exit_code=True):
        out, _ = self.ssh.execute(commands, self._ssh_timeout, check_exit_code)
        return out


def xml_request(check_object=False, check_invalid_data_mover=False):
    """ indicate the return value is a xml api request

    :param check_invalid_data_mover:
    :param check_object:
    :return: the response of this request
    """

    def decorator(f):
        @functools.wraps(f)
        def func_wrapper(self, *argv, **kwargs):
            request = f(self, *argv, **kwargs)
            return self.request(
                request, check_object=check_object,
                check_invalid_data_mover=check_invalid_data_mover)

        return func_wrapper

    return decorator


xml_get_request = xml_request()
xml_set_request = xml_request(check_invalid_data_mover=True)


def nas_command(f):
    """ indicate it's a command of nas command run with ssh

    :param f: function that returns the command in list
    :return: command execution result
    """

    @functools.wraps(f)
    def func_wrapper(self, *argv, **kwargs):
        commands = f(self, *argv, **kwargs)
        return self.ssh_execute(['env', 'NAS_DB=/nas'] + commands)

    return func_wrapper


class VNXNasClient(VNXNasConnections):
    @xml_get_request
    def get_filesystem(self, name=None, fs_id=None):
        return NasXmlBuilder.get_filesystem(name, fs_id=fs_id)

    @xml_set_request
    def create_filesystem(self, name, size, pool_id,
                          mover_id, is_vdm=False):
        return NasXmlBuilder.create_filesystem(
            name, size, pool_id, mover_id, is_vdm)

    @xml_set_request
    def delete_filesystem(self, fs_id):
        return NasXmlBuilder.delete_filesystem(fs_id)

    @xml_set_request
    def extend_fs(self, fs_id, delta_size, pool_id):
        return NasXmlBuilder.extend_filesystem(fs_id, delta_size, pool_id)

    @xml_get_request
    def get_nas_pool(self):
        return NasXmlBuilder.get_nas_pool()

    @xml_get_request
    def get_mover(self, mover_id=None, full=True):
        return NasXmlBuilder.get_mover(mover_id, full)

    @xml_set_request
    def create_dns_domain(self, mover_id, domain_name, servers,
                          protocol='udp'):
        return NasXmlBuilder.create_dns_domain(
            mover_id, domain_name, servers, protocol)

    @xml_set_request
    def delete_dns_domain(self, mover_id, domain_name):
        return NasXmlBuilder.delete_dns_domain(mover_id, domain_name)

    @xml_get_request
    def get_fs_snap(self, name=None, snap_id=None):
        return NasXmlBuilder.get_fs_snap(name, snap_id)

    @xml_set_request
    def create_snap(self, name, fs_id, pool_id, size=None):
        return NasXmlBuilder.create_snap(name, fs_id, pool_id, size)

    @xml_set_request
    def delete_snap(self, snap_id, force=False):
        return NasXmlBuilder.delete_snap(snap_id, force)

    @xml_get_request
    def get_cifs_server(self, name=None, mover_id=None, is_vdm=False):
        return NasXmlBuilder.get_cifs_server(name, mover_id, is_vdm)

    @xml_set_request
    def create_cifs_server(self, name,
                           mover_id, is_vdm=False,
                           workgroup=None, domain=None,
                           ip_list=None,
                           alias_name=None,
                           local_admin_password=None):
        return NasXmlBuilder.create_cifs_server(
            name=name, mover_id=mover_id, is_vdm=is_vdm,
            workgroup=workgroup, domain=domain,
            ip_list=ip_list, alias_name=alias_name,
            local_admin_password=local_admin_password)

    @xml_get_request
    def modify_domain_cifs_server(self, name, mover_id, is_vdm=False,
                                  join_domain=None, username=None,
                                  password=None):
        return NasXmlBuilder.modify_domain_cifs_server(
            name, mover_id, is_vdm, join_domain, username, password)

    @xml_set_request
    def delete_cifs_server(self, name, mover_id=None, is_vdm=False):
        return NasXmlBuilder.delete_cifs_server(name, mover_id, is_vdm)

    @xml_get_request
    def get_fs_mp(self, path=None, mover_id=None, is_vdm=False):
        return NasXmlBuilder.get_fs_mp(path, mover_id, is_vdm)

    @xml_set_request
    def create_fs_mp(self, path, fs_id, mover_id, is_vdm=False):
        return NasXmlBuilder.create_fs_mp(path, fs_id, mover_id, is_vdm)

    @xml_set_request
    def delete_fs_mp(self, path, mover_id, is_vdm=False):
        return NasXmlBuilder.delete_fs_mp(path, mover_id, is_vdm)

    @xml_get_request
    def get_mover_host(self, mover_host_id=None):
        return NasXmlBuilder.get_mover_host(mover_host_id)

    @xml_set_request
    def create_mover_interface(self, mover_id, device, ip, net_mask,
                               vlan_id=0, name=None):
        return NasXmlBuilder.create_mover_interface(
            mover_id, device, ip, net_mask, vlan_id, name)

    @xml_set_request
    def delete_mover_interface(self, mover_id, ip):
        return NasXmlBuilder.delete_mover_interface(mover_id, ip)

    @nas_command
    def get_mover_interconnect_id_list(self):
        return NasCommand.nas_cel_list()

    @xml_get_request
    def get_vdm(self, vdm_id=None):
        return NasXmlBuilder.get_vdm(vdm_id)

    @xml_set_request
    def create_vdm(self, mover_id, name, pool_id=None):
        return NasXmlBuilder.create_vdm(mover_id, name, pool_id)

    @xml_set_request
    def delete_vdm(self, vdm_id):
        return NasXmlBuilder.delete_vdm(vdm_id)

    @nas_command
    def get_dm_interfaces(self, name=None, is_vdm=True):
        return NasCommand.get_dm_interfaces(name, is_vdm)

    @nas_command
    def attach_nfs_interface(self, if_name, vdm_name=None):
        return NasCommand.attach_nfs_interface(if_name=if_name,
                                               vdm_name=vdm_name)

    @nas_command
    def detach_nfs_interface(self, if_name, vdm_name=None):
        return NasCommand.detach_nfs_interface(if_name=if_name,
                                               vdm_name=vdm_name)

    @xml_get_request
    def get_nfs_export(self, mover_id=None, path=None):
        return NasXmlBuilder.get_nfs_export(mover_id, path)

    @xml_set_request
    def create_nfs_export(self, mover_id, path, ro=False, host_config=None):
        return NasXmlBuilder.create_nfs_export(mover_id, path, ro,
                                               host_config)

    @xml_set_request
    def delete_nfs_export(self, mover_id, path):
        return NasXmlBuilder.delete_nfs_export(mover_id, path)

    @xml_set_request
    def modify_nfs_export(self, mover_id, path, ro=None, host_config=None):
        return NasXmlBuilder.modify_nfs_export(mover_id, path, ro,
                                               host_config)

    @xml_get_request
    def get_cifs_share(self, server_name=None, share_name=None,
                       mover_id=None, is_vdm=False):
        return NasXmlBuilder.get_cifs_share(
            server_name, share_name, mover_id, is_vdm)

    @xml_set_request
    def create_cifs_share(self, name, server_name, mover_id,
                          is_vdm=False, path=None):
        return NasXmlBuilder.create_cifs_share(name, server_name, mover_id,
                                               is_vdm, path)

    @xml_set_request
    def delete_cifs_share(self, name, mover_id, server_names, is_vdm=False):
        return NasXmlBuilder.delete_cifs_share(
            name=name, mover_id=mover_id, server_names=server_names,
            is_vdm=is_vdm)

    @nas_command
    def disable_cifs_share_access(self, share_name, mover_name):
        return NasCommand.disable_cifs_share_access(share_name, mover_name)

    @nas_command
    def allow_cifs_share_access(self, share_name, mover_name, user_name,
                                domain, access=CifsAccessControl.FULL):
        return NasCommand.allow_cifs_share_access(
            share_name, mover_name, user_name, domain, access)

    @nas_command
    def deny_cifs_share_access(self, share_name, mover_name, user_name,
                               domain, access=CifsAccessControl.FULL):
        return NasCommand.deny_cifs_share_access(
            share_name, mover_name, user_name, domain, access)


class XmlStatus(Enum):
    OK = 'ok'
    INFO = 'info'
    DEBUG = 'debug'
    WARNING = 'warning'
    ERROR = 'error'


class NasXmlResponse(object):
    def __init__(self, resp, parser=None):
        self._check_credential_error(resp)
        resp = self._parse_resp(parser, resp)
        self._dict = resp

    @staticmethod
    def _check_credential_error(resp):
        if 'Session timeout. Relogin and try this operation again.' in resp:
            raise VNXFileCredentialError()

    @staticmethod
    def _parse_resp(parser, resp):
        if parser is None:
            parser = XMLAPIParser()
        if isinstance(resp, six.string_types):
            resp = parser.parse(resp)
        return resp

    @property
    def status(self):
        return XmlStatus.parse(self._dict['maxSeverity'])

    @status.setter
    def status(self, value):
        self._dict['maxSeverity'] = value

    @property
    def problems(self):
        return self._dict['problems']

    @property
    def objects(self):
        return self._dict['objects']

    def filter_object(self, **kwargs):
        objects = self.objects
        for k, v in kwargs.items():
            if v is not None:
                v = str(v)
                objects = list(filter(lambda obj: obj[k] == v, objects))
        self._dict['objects'] = objects

    @property
    def first_object(self):
        if len(self.objects) < 1:
            raise ValueError('not a single object available.')
        return self.objects[0]

    @property
    def problem_message_codes(self):
        return set(map(int, self._get_problem_props('messageCode')))

    @property
    def hex_problem_message_codes(self):
        ret = self.problem_message_codes
        return set(map(to_hex, ret))

    @property
    def problem_messages(self):
        return '\n'.join(set(self._get_problem_props('message')))

    @property
    def problem_diagnostics(self):
        return '\n'.join(set(self._get_problem_props('Diagnostics')))

    def _get_problem_props(self, key):
        return [problem[key] for problem in self.problems
                if key in problem]

    def get_status_msg(self, desc=None):
        msg = ('status: {}.\n'
               'problem details: \n'
               '{}'
               .format(self.status,
                       self.get_problems_string(' ' * 4)))
        if desc is not None:
            msg = '  '.join([desc, msg])
        return msg

    def get_problems_string(self, prefix=''):
        msgs = []
        for i, v in enumerate(self.problems, 1):
            msgs.append('{p}({i}) code: {mc}\n'
                        '{p}    message: {msg}\n'
                        '{p}    diagnostics: {d}\n'
                        .format(p=prefix, i=i,
                                mc=v['messageCode'],
                                msg=v['message'],
                                d=v.get('Diagnostics', 'N/A')))
        return ''.join(msgs)

    def check_invalid_data_mover(self):
        if not self.is_ok():
            try:
                self.raise_if_err()
            except VNXInvalidMoverID:
                raise
            except VNXException:
                # pass to upper level for further process
                pass

    def raise_if_err(self, desc=None):
        if not self.is_ok():
            msg = self.get_status_msg(desc)
            exception_clz = get_xmlapi_exception(
                self.problem_message_codes, default=VNXBackendError)
            raise exception_clz(message=msg)

    def raise_if_no_object(self, desc=None):
        if not self.objects:
            msg = self.get_status_msg(desc)
            raise VNXNasObjectNotFound(err=msg)

    def is_ok(self):
        return self.status in (XmlStatus.DEBUG, XmlStatus.INFO, XmlStatus.OK)

    def is_error(self):
        return self.status == XmlStatus.ERROR

    def has_error_code(self, code):
        return to_int(code) in self.problem_message_codes



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
from unittest import TestCase

from hamcrest import equal_to, assert_that, not_none, none, raises

from storops.vnx.enums import VNXSPEnum
from storops.vnx.parsers import VNXCliParser, VNXPropDescriptor, \
    VNXParserConfigFactory
from storops.vnx.resource import get_vnx_parser
from storops_test.vnx.cli_mock import MockCli
from storops_test.vnx.resource.fakes import STORAGE_GROUP_HBA

log = logging.getLogger(__name__)

A = VNXPropDescriptor('-a', 'Prop A (name):', 'prop_a')
B = VNXPropDescriptor('-b', 'Prop B:')
C = VNXPropDescriptor('-c', 'Prop C:')
ID = VNXPropDescriptor(None, 'ID:', is_index=True)


class DemoParser(VNXCliParser):
    def __init__(self):
        super(DemoParser, self).__init__()
        self.add_property(A, B, C, ID)


class DemoParserNonIndex(VNXCliParser):
    def __init__(self):
        super(DemoParserNonIndex, self).__init__()
        self.add_property(VNXPropDescriptor('-b', 'Prop B:'))


class DemoParserRegexIndex(VNXCliParser):
    def __init__(self):
        super(DemoParserRegexIndex, self).__init__()
        self.add_property(
            VNXPropDescriptor(None,
                              r'\s*\w+:(\d+)',
                              'id',
                              is_index=True,
                              is_regex=True,
                              converter=int),
            VNXPropDescriptor(None,
                              r'\s*value:\s*(\w+)',
                              'value',
                              is_regex=True))


class DemoParserMultiIndices(VNXCliParser):
    def __init__(self):
        super(DemoParserMultiIndices, self).__init__()
        self.add_property(
            VNXPropDescriptor(None, 'A:', is_index=True),
            VNXPropDescriptor(None, 'B:', is_index=True),
            VNXPropDescriptor(None, 'C:'),
            VNXPropDescriptor(None, 'D:'))


class VNXCliParserTest(TestCase):
    def test_get_property_options(self):
        options = DemoParser().property_options
        assert_that(' '.join(options), equal_to('-a -b -c'))

    def test_get_index_descriptor(self):
        assert_that(DemoParser().index_property.label, equal_to('ID:'))

    def test_get_index_descriptor_none(self):
        assert_that(DemoParserNonIndex().index_property, none())

    def test_parse(self):
        output = """
                ID: test
                Prop A (Name): ab (c)
                Prop B: d ef
                """
        parser = DemoParser()
        parsed = parser.parse(output, [A, ID, C])

        assert_that(parsed.prop_a, equal_to('ab (c)'))
        assert_that(parsed.prop_c, none())
        assert_that(parsed.id, equal_to('test'))

        def f():
            log.debug(parsed.prop_b)

        assert_that(f, raises(AttributeError))

    def test_parse_empty_prop(self):
        output = """
                ID: test
                Prop A (Name): ab (c)
                Prop B:
                Prop C: abc
                """
        parser = DemoParser()
        parsed = parser.parse(output, [A, ID, B, C])

        assert_that(parsed.id, equal_to('test'))
        assert_that(parsed.prop_a, equal_to('ab (c)'))
        assert_that(parsed.prop_b, equal_to(''))

    def test_parse_regex_label(self):
        output = """
                id:123
                value:abcde
                id:456
                value:ghijk
                """
        parsed = DemoParserRegexIndex().parse_all(output)
        assert_that(len(parsed), equal_to(2))
        for i in parsed:
            if i.id == 123:
                assert_that(i.value, equal_to('abcde'))
            elif i.id == 456:
                assert_that(i.value, equal_to('ghijk'))
            else:
                self.fail('id not recognized.')

    def test_all_options(self):
        options = DemoParser().all_options
        assert_that(options, equal_to(['-a', '-b', '-c']))

    def test_parse_multi_index(self):
        output = """
        A: a0
        B: b0
        C: c0

        A: a0
        B: b0
        D: d0

        A: a0
        B: b1
        C: c1
        """
        parsed = DemoParserMultiIndices().parse_all(output)
        assert_that(len(parsed), equal_to(2))
        a0b0 = next(i for i in parsed if i.b == 'b0')
        assert_that(a0b0, not_none())
        assert_that(a0b0.a, equal_to('a0'))
        assert_that(a0b0.b, equal_to('b0'))
        assert_that(a0b0.c, equal_to('c0'))
        assert_that(a0b0.d, equal_to('d0'))

        a0b1 = next(i for i in parsed if i.b == 'b1')
        assert_that(a0b1, not_none())
        assert_that(a0b1.a, equal_to('a0'))
        assert_that(a0b1.b, equal_to('b1'))
        assert_that(a0b1.c, equal_to('c1'))


class VNXStorageGroupHBAParserTest(TestCase):
    def test_parse(self):
        data = get_vnx_parser("VNXStorageGroupHBA").parse(STORAGE_GROUP_HBA)
        assert_that(data.host_name, equal_to('abc.def.dev'))
        assert_that(data.sp_port, equal_to('A-3v1'))
        assert_that(data.initiator_ip, equal_to('10.244.209.72'))
        assert_that(data.tpgt, equal_to('1'))
        assert_that(data.isid, equal_to('10000000000'))
        assert_that(
            data.hba,
            equal_to(('iqn.1991-05.com.microsoft:abc.def.dev',
                      'SP A', '3')))

    def test_parse_no_header(self):
        output = """
                iqn.1991-05.com.microsoft:abc.def.dev  SP A     1
                Host name:             abc.def.dev
                SPPort:                A-1v0
                Initiator IP:          10.244.209.72
                TPGT:                  1
                ISID:                  10000000000
                """
        data = get_vnx_parser("VNXStorageGroupHBA").parse(output)
        assert_that(data.host_name, equal_to('abc.def.dev'))
        assert_that(data.sp_port, equal_to('A-1v0'))
        assert_that(data.initiator_ip, equal_to('10.244.209.72'))
        assert_that(data.tpgt, equal_to('1'))
        assert_that(data.isid, equal_to('10000000000'))
        assert_that(data.hba,
                    equal_to(('iqn.1991-05.com.microsoft:abc.def.dev',
                              'SP A',
                              '1')))


class VNXStorageGroupParserTest(TestCase):
    def test_parse(self):
        parser = get_vnx_parser('VNXStorageGroup')
        output = MockCli.read_file('storagegroup_-messner_-list_-host_'
                                   '-iscsiAttributes_-gname_microsoft.txt')
        sg = parser.parse(output)
        assert_that(sg.shareable, equal_to(True))
        assert_that(sg.name, equal_to('microsoft'))
        assert_that(
            sg.wwn,
            equal_to('12:34:56:78:9A:BC:DE:F1:23:45:67:89:AB:CD:EF:01'))
        assert_that(sg.alu_hlu_map[4], equal_to(0))
        assert_that(sg.alu_hlu_map[456], equal_to(123))
        assert_that(sg.alu_hlu_map.get(3, None), none())

        # assert for hba members
        assert_that(len(sg.hba_sp_pairs), equal_to(2))
        hba = sg.hba_sp_pairs[0]
        assert_that(hba.host_name, equal_to('abc.def.dev'))


class VNXConsistencyGroupParserTest(TestCase):
    def test_parse(self):
        output = MockCli.read_file('snap_-group_-list_-detail.txt')
        parser = get_vnx_parser('VNXConsistencyGroup')
        cgs = parser.parse_all(output)
        cg = next(c for c in cgs if c.name == 'test cg name')
        assert_that(cg, not_none())
        assert_that(cg.state, equal_to('Ready'))

        cg = next(c for c in cgs if c.name == 'another cg')
        assert_that(cg, not_none())
        assert_that(cg.state, equal_to('Offline'))


class VNXPoolPropertiesTest(TestCase):
    def test_parse(self):
        output = MockCli.read_file('storagepool_-list_-all_-id_1.txt')
        parser = get_vnx_parser('VNXPool')
        pool = parser.parse(output)
        assert_that(pool.state, equal_to('Ready'))
        assert_that(pool.pool_id, equal_to(1))
        assert_that(pool.user_capacity_gbs, equal_to(2329.792))
        assert_that(pool.available_capacity_gbs, equal_to(1473.623))
        assert_that(pool.fast_cache, none())
        assert_that(pool.name, equal_to('Pool_daq'))
        assert_that(pool.total_subscribed_capacity_gbs, equal_to(2701.767))
        assert_that(pool.percent_full_threshold, equal_to(70))


class VNXPoolFeatureParserTest(TestCase):
    # command:  storagepool -feature -info
    output = """
    Is Virtual Provisioning Supported:  true
    Max. Pools:  60
    Max. Disks Per Pool:  1496
    Max. Disks for all Pools:  1496
    Max. Disks Per Operation:  180
    Max. Pool LUNs:  4000
    Min. Pool LUN Size(Blocks):  1
    Max. Pool LUN Size(Blocks):  549755813888
    Max. Pool LUN Size(GBs):  262144.000
    Total Number of Pools:  2
    Total Number of Pool LUNs:  4
    Total Number of all Pool LUNs that are thin:  3
    Total Number of all Pool LUNs that are non-thin:  1
    Number of Disks used in Pools:  5
    Available Disks:
    Bus 0 Enclosure 0 Disk 24
    Bus 0 Enclosure 0 Disk 16
    Bus 0 Enclosure 0 Disk 5
    Bus 0 Enclosure 0 Disk 4
    """

    def test_parse(self):
        parser = get_vnx_parser('VNXPoolFeature')
        parsed = parser.parse(self.output)
        assert_that(parsed.max_pool_luns, equal_to(4000))
        assert_that(parsed.total_pool_luns, equal_to(4))


class VNXLunPropertiesTest(TestCase):
    def test_parse(self):
        output = MockCli.read_file('lun_-list_-all_-l_19.txt')
        parser = get_vnx_parser('VNXLun')
        parsed = parser.parse(output)
        wwn = '60:06:01:60:1A:50:35:00:CC:22:61:D6:76:B1:E4:11'
        assert_that(parsed.wwn, equal_to(wwn))
        assert_that(parsed.name, equal_to('test_lun'))
        assert_that(parsed.lun_id, equal_to(19))
        assert_that(parsed.total_capacity_gb, equal_to(1.0))
        assert_that(parsed.is_thin_lun, equal_to(True))
        assert_that(parsed.is_compressed, equal_to(False))
        assert_that(parsed.deduplication_state, equal_to('Off'))
        assert_that(parsed.tiering_policy, equal_to('No Movement'))
        assert_that(parsed.initial_tier, equal_to('Optimize Pool'))
        assert_that(parsed.state, equal_to('Ready'))
        assert_that(parsed.status, equal_to('OK(0x0)'))
        assert_that(parsed.operation, equal_to('None'))
        assert_that(parsed.current_owner, equal_to(VNXSPEnum.SP_A))
        assert_that(parsed.attached_snapshot, none())


class VNXParserConfigFactoryTest(TestCase):
    def test_read_properties(self):
        name = 'VNXConsistencyGroup'
        prop = get_vnx_parser(name)
        assert_that(prop.resource_class_name, equal_to(name))
        assert_that(prop.data_src, equal_to('cli'))

    def test_properties_sequence_should_align_with_file(self):
        props = get_vnx_parser('VNXSystem')
        assert_that(props.MODEL.sequence, equal_to(0))
        assert_that(props.NAME.sequence, equal_to(5))

    def test_get_rsc_pkg_name(self):
        name = VNXParserConfigFactory.get_rsc_pkg_name()
        assert_that(name, equal_to('storops.vnx.resource'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

from hamcrest import assert_that, greater_than_or_equal_to, \
    raises, contains_string
from hamcrest import equal_to

from storops_test.vnx.nas_mock import t_nas, patch_post
from storops.exception import VNXBackendError
from storops.vnx.resource.cifs_server import VNXCifsServer, CifsDomain

__author__ = 'Jay Xu'


class VNXCifsServerTest(unittest.TestCase):
    # todo: test modify

    @patch_post
    def test_get_all(self):
        cifs_list = VNXCifsServer.get(t_nas())
        assert_that(len(cifs_list), greater_than_or_equal_to(1))
        cifs = next(cifs for cifs in cifs_list if cifs.name == 'CIFS')
        self.verify_pie_cifs(cifs)

    @patch_post
    def test_get_all_by_mover(self):
        cifs_list = VNXCifsServer.get(t_nas(), mover_id=1)
        cifs = next(cifs for cifs in cifs_list if cifs.name == 'CIFS')
        self.verify_pie_cifs(cifs)

    @patch_post
    def test_get_all_by_mover_not_found(self):
        cifs_list = VNXCifsServer.get(t_nas(), mover_id=1, is_vdm=True)
        assert_that(len(cifs_list), equal_to(0))

    @patch_post
    def test_get_by_name(self):
        cifs = VNXCifsServer.get(t_nas(), 'CIFS')
        self.verify_pie_cifs(cifs)

    @patch_post
    def test_get_by_name_not_found(self):
        cifs = VNXCifsServer('not_found', t_nas())
        assert_that(cifs.existed, equal_to(False))
        assert_that(cifs._get_name(), equal_to('not_found'))

    def verify_pie_cifs(self, cifs):
        assert_that(cifs.mover_id, equal_to(1))
        assert_that(cifs.existed, equal_to(True))
        assert_that(cifs.name, equal_to('CIFS'))
        assert_that(cifs.interfaces, equal_to('10.110.24.194'))
        assert_that(cifs.is_vdm, equal_to(False))
        assert_that(cifs.workgroup, equal_to('PIE'))
        assert_that(cifs.local_users, equal_to(True))
        assert_that(cifs.type, equal_to('standalone'))

    @patch_post
    def test_create_cifs_server_invalid_domain_name(self):
        try:
            domain = CifsDomain('test_domain')
            VNXCifsServer.create(t_nas(), 'test', 1, domain=domain)
            self.fail('should raise an exception.')
        except VNXBackendError as ex:
            assert_that(ex.message, contains_string('not facet-valid'))

    @patch_post
    def test_create_cifs_server_no_default_nt_server(self):
        def f():
            domain = CifsDomain('test.dev')
            VNXCifsServer.create(t_nas(), 'test', 1, domain=domain)

        assert_that(f, raises(VNXBackendError, 'default NT server'))

    @patch_post
    def test_create_cifs_server_net_bios_existed(self):
        def f():
            domain = CifsDomain('test.dev')
            VNXCifsServer.create(t_nas(), 'test', 1,
                                 domain=domain,
                                 interfaces='10.110.24.194')

        assert_that(f, raises(VNXBackendError, 'already exists as server'))

    @patch_post
    def test_create_cifs_server_w2k(self):
        domain = CifsDomain('test.dev')
        cifs = VNXCifsServer.create(t_nas(), 's8', 1, domain=domain,
                                    interfaces='10.110.24.194')
        assert_that(cifs.domain, equal_to('TEST.DEV'))
        assert_that(cifs.type, equal_to('W2K'))
        assert_that(cifs.comp_name, equal_to('S8'))
        assert_that(cifs.domain_joined, equal_to(False))

    @patch_post
    def test_delete_cifs_server(self):
        cifs = VNXCifsServer('test', t_nas())
        resp = cifs.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_post
    def test_delete_cifs_server_not_found(self):
        def f():
            cifs = VNXCifsServer('test1', t_nas())
            cifs.delete(1)

        assert_that(f, raises(VNXBackendError, 'does not exist'))

    @patch_post
    def test_create_cifs_server_stand_alone(self):
        cifs = VNXCifsServer.create(t_nas(), 's2', 1, workgroup='work',
                                    interfaces='10.110.24.194',
                                    local_admin_password='password')
        assert_that(cifs.workgroup, equal_to('WORK'))
        assert_that(cifs.name, equal_to('S2'))

    @patch_post
    def test_create_cifs_default_mover(self):
        cifs = VNXCifsServer.create(t_nas(), 's2', workgroup='work',
                                    interfaces='10.110.24.194',
                                    local_admin_password='password')
        assert_that(cifs.existed, equal_to(True))
        assert_that(cifs.workgroup, equal_to('WORK'))
        assert_that(cifs.name, equal_to('S2'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

import ddt
from hamcrest import assert_that, equal_to, instance_of, only_contains, \
    raises, contains_string

from storops.exception import SystemAPINotSupported
from storops.exception import UnityEthernetPortSpeedNotSupportError, \
    UnityEthernetPortMtuSizeNotSupportError, UnityResourceNotFoundError, \
    UnityPolicyNameInUseError, UnityPolicyIncorrectRequestBodyError, \
    UnityEthernetPortAlreadyAggregatedError, UnityIPInUseException
from storops.unity.enums import EPSpeedValuesEnum, \
    FcSpeedEnum, IOLimitPolicyStateEnum, IOLimitPolicyTypeEnum, \
    SpeedValuesEnum, ConnectorTypeEnum
from storops.unity.resource.health import UnityHealth
from storops.unity.resource.lun import UnityLun
from storops.unity.resource.port import UnityEthernetPort, \
    UnityEthernetPortList, UnityIpPort, UnityIpPortList, UnityIscsiPortal, \
    UnityIscsiPortalList, UnityIscsiNode, UnityFcPort, UnityFcPortList, \
    UnityIoLimitRule, UnityIoLimitPolicy, UnityIoLimitPolicyList, \
    UnityLinkAggregation, UnityLinkAggregationList, UnitySasPort
from storops.unity.resource.sp import UnityStorageProcessor
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'


class UnityIpPortTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        port = UnityIpPort('spa_eth2', cli=t_rest())
        assert_that(port.name, equal_to('SP A Ethernet Port 2'))
        assert_that(port.short_name, equal_to('Ethernet Port 2'))
        assert_that(port.sp, instance_of(UnityStorageProcessor))
        assert_that(port.is_link_up, equal_to(True))
        assert_that(port.mac_address, equal_to('00:60:16:5C:08:E1'))

    @patch_rest
    def test_get_all(self):
        ports = UnityIpPortList(cli=t_rest())
        assert_that(len(ports), equal_to(8))

    @patch_rest
    def test_is_link_aggregation(self):
        port = UnityIpPort('spa_eth3', cli=t_rest())
        assert_that(port.is_link_aggregation(), equal_to(False))
        port = UnityIpPort('spa_la_2', cli=t_rest())
        assert_that(port.is_link_aggregation(), equal_to(True))

    @patch_rest
    def test_is_link_aggregation_not_supported(self):
        port = UnityIpPort('spa_eth6', cli=t_rest("4.0"))
        assert_that(port.is_link_aggregation(), equal_to(False))

    @patch_rest
    def test_set_mtu_on_eth_port(self):
        port = UnityIpPort('spa_eth3', cli=t_rest())
        port.set_mtu(1500)

    @patch_rest
    def test_set_mtu_on_link_aggregation(self):
        la = UnityIpPort('spa_la_2', cli=t_rest())
        la.set_mtu(1500)

    @patch_rest
    def test_mtu_of_link_aggregation(self):
        la = UnityIpPort('spa_la_2', cli=t_rest())
        assert_that(la.mtu, equal_to(9000))

    @patch_rest
    def test_mtu_of_eth_port(self):
        port = UnityIpPort('spa_eth3', cli=t_rest())
        assert_that(port.mtu, equal_to(1500))


@ddt.ddt
class UnityEthernetPortTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        port = UnityEthernetPort('spa_eth3', cli=t_rest())
        assert_that(port.name, equal_to('SP A Ethernet Port 3'))
        assert_that(port.mac_address, equal_to("00:60:16:5C:07:0A"))
        assert_that(port.parent_storage_processor, equal_to(
            UnityStorageProcessor('spa', cli=t_rest())))
        assert_that(port.mtu, equal_to(1500))
        assert_that(port.requested_mtu, equal_to(1500))
        assert_that(port.connector_type, equal_to(ConnectorTypeEnum.RJ45))
        assert_that(port.supported_speeds, only_contains(
            EPSpeedValuesEnum.AUTO,
            EPSpeedValuesEnum._100MbPS,
            EPSpeedValuesEnum._1GbPS,
            EPSpeedValuesEnum._10GbPS))
        assert_that(port.supported_mtus, only_contains(1500, 9000))
        assert_that(port.speed, equal_to(None))
        assert_that(port.needs_replacement, equal_to(False))
        assert_that(port.is_link_up, equal_to(False))
        assert_that(port.bond, equal_to(False))

    @patch_rest
    def test_get_properties_io_module_port(self):
        port = UnityEthernetPort('spa_iom_0_eth0', cli=t_rest())
        assert_that(port.name, equal_to('SP A I/O Module 0 Ethernet Port 0'))
        assert_that(port.mac_address, equal_to("00:60:16:57:EC:58"))
        assert_that(port.parent_storage_processor, equal_to(
            UnityStorageProcessor('spa', cli=t_rest())))
        assert_that(port.mtu, equal_to(1500))
        assert_that(port.requested_mtu, equal_to(0))
        assert_that(port.connector_type, equal_to(ConnectorTypeEnum.LC))
        assert_that(port.supported_speeds, only_contains(
            EPSpeedValuesEnum.AUTO,
            EPSpeedValuesEnum._100MbPS,
            EPSpeedValuesEnum._1GbPS,
            EPSpeedValuesEnum._10GbPS))
        assert_that(port.supported_mtus, only_contains(1500, 9000))
        assert_that(port.speed, equal_to(EPSpeedValuesEnum._10GbPS))
        assert_that(port.needs_replacement, equal_to(False))
        assert_that(port.is_link_up, equal_to(True))
        assert_that(port.bond, equal_to(False))

    @patch_rest
    def test_modify_mtu(self):
        port = UnityEthernetPort(cli=t_rest(), _id='spa_eth3')
        port.modify(mtu=9000)

    @patch_rest
    def test_modify_mtu_to_invalid_value(self):
        def do():
            port = UnityEthernetPort(cli=t_rest(), _id='spb_eth3')
            port.modify(mtu=10000)

        assert_that(do, raises(UnityEthernetPortMtuSizeNotSupportError))

    @patch_rest
    def test_modify_mtu_to_same_value(self):
        port = UnityEthernetPort(cli=t_rest(), _id='spb_eth3')
        port.modify(mtu=1500)

    @patch_rest
    def test_modify_speed(self):
        port = UnityEthernetPort(cli=t_rest(), _id='spa_eth3')
        port.modify(speed=100)

    @patch_rest
    def test_modify_speed_to_same_value(self):
        port = UnityEthernetPort(cli=t_rest(), _id='spb_eth3')
        port.modify(speed=EPSpeedValuesEnum.AUTO)

    @patch_rest
    def test_modify_speed_to_invalid_value(self):
        def do():
            port = UnityEthernetPort(cli=t_rest(), _id='spb_eth3')
            port.modify(speed=40000)

        assert_that(do, raises(UnityEthernetPortSpeedNotSupportError))

    @patch_rest
    def test_modify_when_peer_not_exist(self):
        port = UnityEthernetPort(cli=t_rest(), _id='spa_eth4')
        port.modify(mtu=1500)

    @ddt.data({'port_id': 'spa_eth2',
               'peer_id': 'spb_eth2'},
              {'port_id': 'spb_eth3',
               'peer_id': 'spa_eth3'})
    @ddt.unpack
    def test_get_peer(self, port_id, peer_id):
        port = UnityEthernetPort(cli=t_rest(), _id=port_id)
        peer = port.get_peer()
        assert_that(peer.get_id(), equal_to(peer_id))


class UnityEthernetPortListTest(TestCase):
    @patch_rest
    def test_filter(self):
        port_id = 'spa_eth2'

        all_ports = UnityEthernetPortList(cli=t_rest())
        ports = all_ports.shadow_copy(port_ids=[port_id])
        assert_that(len(ports), equal_to(1))
        assert_that(ports[0].get_id(), equal_to(port_id))

        ports = UnityEthernetPortList(cli=t_rest(), port_ids=[port_id])
        assert_that(len(ports), equal_to(1))
        assert_that(ports[0].get_id(), equal_to(port_id))


class UnityIscsiPortalTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        portal = UnityIscsiPortal(cli=t_rest(), _id='if_4')
        assert_that(portal.ip_address, equal_to('10.244.213.177'))
        assert_that(portal.iscsi_node, instance_of(UnityIscsiNode))
        assert_that(portal.iscsi_node.name,
                    equal_to('iqn.1992-04.com.emc:cx.fnm00150600267.a0'))
        assert_that(portal.netmask, equal_to('255.255.255.0'))
        assert_that(portal.gateway, equal_to('10.244.213.1'))
        assert_that(portal.ethernet_port, instance_of(UnityEthernetPort))

    @patch_rest
    def test_create_iscsi_portal(self):
        portal = UnityIscsiPortal.create(
            cli=t_rest(), ethernet_port='spa_eth3', ip="10.244.213.244",
            netmask="255.255.255.0", vlan=133, gateway="10.244.213.1")
        assert_that(portal.id, equal_to("if_4"))

    @patch_rest
    def test_create_duplicate_portal(self):
        def f():
            UnityIscsiPortal.create(
                cli=t_rest(), ethernet_port='spa_eth2', ip='10.244.213.244',
                netmask="255.255.255.0", gateway="10.244.213.1")

        assert_that(f, raises(UnityIPInUseException))

    @patch_rest
    def test_modify_portal(self):
        portal = UnityIscsiPortal(cli=t_rest(), _id='if_4')
        portal.modify(ip="10.244.21.11", netmask="255.255.252.0",
                      vlan=111, gateway="10.244.21.1")

    @patch_rest
    def test_delete_portal(self):
        portal = UnityIscsiPortal.get(cli=t_rest(), _id='if_4')
        portal.delete()

    @patch_rest
    def test_delete_portal_not_found(self):
        portal = UnityIscsiPortal.get(cli=t_rest(), _id='if_20')

        def f():
            portal.delete()

        assert_that(f, raises(UnityResourceNotFoundError))


class UnityIscsiPortalListTest(TestCase):
    @patch_rest
    def test_filter(self):
        port_id = 'spa_eth2'

        all_ports = UnityIscsiPortalList(cli=t_rest())
        ports = all_ports.shadow_copy(port_ids=[port_id])
        assert_that(len(ports), equal_to(2))
        assert_that(set(ports.id), equal_to(set(['if_4', 'if_5'])))

        ports = UnityIscsiPortalList(cli=t_rest(), port_ids=[port_id])
        assert_that(len(ports), equal_to(2))
        assert_that(set(ports.id), equal_to(set(['if_4', 'if_5'])))


class UnityFcPortTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        port = UnityFcPort('spa_fc4', cli=t_rest())
        assert_that(port.existed, equal_to(True))
        assert_that(port.slot_number, equal_to(4))
        assert_that(
            port.wwn,
            equal_to("50:06:01:60:C7:E0:01:DA:50:06:01:62:47:E0:01:DA"))
        assert_that(port.available_speeds,
                    only_contains(FcSpeedEnum._4GbPS,
                                  FcSpeedEnum._8GbPS,
                                  FcSpeedEnum._16GbPS,
                                  FcSpeedEnum.AUTO))
        assert_that(port.connector_type,
                    equal_to(ConnectorTypeEnum.LC))
        assert_that(port.name,
                    equal_to("SP A FC Port 4"))
        assert_that(port.storage_processor,
                    equal_to(UnityStorageProcessor('spa', cli=t_rest())))


class UnityFcPortListTest(TestCase):
    @patch_rest
    def test_filter(self):
        port_id = 'spa_iom_1_fc2'

        all_ports = UnityFcPortList(cli=t_rest())
        ports = all_ports.shadow_copy(port_ids=[port_id])
        assert_that(len(ports), equal_to(1))
        assert_that(ports[0].get_id(), equal_to(port_id))

        ports = UnityFcPortList(cli=t_rest(), port_ids=[port_id])
        assert_that(len(ports), equal_to(1))
        assert_that(ports[0].get_id(), equal_to(port_id))


class UnityIoLimitRuleTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        rule = UnityIoLimitRule('qr_1', cli=t_rest())
        assert_that(rule.get_id(), equal_to('qr_1'))
        assert_that(rule.max_iops, equal_to(1000))
        assert_that(rule.io_limit_policy.get_id(), equal_to('qp_1'))
        assert_that(rule.name, equal_to('Limit_1000_IOPS_rule'))


class UnityIoLimitPolicyTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        rule = UnityIoLimitPolicy('qp_2', cli=t_rest())
        assert_that(rule.description, contains_string('Absolute bandwidth'))
        assert_that(rule.existed, equal_to(True))
        assert_that(rule.get_id(), equal_to('qp_2'))
        assert_that(rule.is_shared, equal_to(False))
        lun_list = rule.luns
        assert_that(len(lun_list), equal_to(1))
        assert_that(rule.name, equal_to('Limit_2_MBPS'))
        assert_that(rule.state, equal_to(IOLimitPolicyStateEnum.ACTIVE))
        assert_that(rule.type, equal_to(IOLimitPolicyTypeEnum.ABSOLUTE))
        settings = rule.io_limit_rule_settings
        assert_that(len(settings), equal_to(1))

        setting = settings[0]
        assert_that(setting.burst_frequency, equal_to(1))
        assert_that(setting.burst_time, equal_to(5))
        assert_that(setting.get_id(), equal_to('qr_2'))
        assert_that(setting.max_kbps, equal_to(2048))
        assert_that(setting.name, equal_to('Limit_2_MBPS_rule'))

    @patch_rest
    def test_get_list(self):
        rule_list = UnityIoLimitPolicyList.get(cli=t_rest())
        assert_that(len(rule_list), equal_to(6))

    @patch_rest
    def test_create_kbps_policy(self):
        policy = UnityIoLimitPolicy.create(
            t_rest(), 'max_kbps_1234', max_kbps=1234, description='storops')
        assert_that(policy.name, equal_to('max_kbps_1234'))
        assert_that(policy.is_paused, equal_to(False))
        setting = policy.io_limit_rule_settings[0]
        assert_that(setting.max_kbps, equal_to(1234))
        assert_that(setting.name, equal_to('max_kbps_1234_rule'))

    @patch_rest
    def test_create_policy_existed(self):
        def f():
            UnityIoLimitPolicy.create(t_rest(), 'test1', max_kbps=1)

        assert_that(f, raises(UnityPolicyNameInUseError, 'been reserved'))

    @patch_rest
    def test_create_policy_with_incorrect_parameters(self):
        def f():
            UnityIoLimitPolicy.create(t_rest(),
                                      'float_type_kbps',
                                      max_kbps=1.0)

        assert_that(f, raises(UnityPolicyIncorrectRequestBodyError,
                              'Request body is not correct'))

    @patch_rest
    def test_create_iops_policy(self):
        policy = UnityIoLimitPolicy.create(
            t_rest(), 'max_iops_4321', max_iops=4321, description='storops')
        assert_that(policy.name, equal_to('max_iops_4321'))
        assert_that(policy.is_paused, equal_to(False))
        setting = policy.io_limit_rule_settings[0]
        assert_that(setting.max_iops, equal_to(4321))
        assert_that(setting.name, equal_to('max_iops_4321_rule'))

    @patch_rest
    def test_delete_policy_not_found(self):
        def f():
            policy = UnityIoLimitPolicy('qp_8', t_rest())
            policy.delete()

        assert_that(f, raises(UnityResourceNotFoundError))

    @patch_rest
    def test_apply_to_storage(self):
        policy = UnityIoLimitPolicy('qp_5', t_rest())
        lun1 = UnityLun('sv_2024', t_rest())
        lun2 = UnityLun('sv_2025', t_rest())
        resp = policy.apply_to_storage(lun1, lun2)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_remove_from_storage(self):
        policy = UnityIoLimitPolicy('qp_5', t_rest())
        lun1 = UnityLun('sv_2024', t_rest())
        lun2 = UnityLun('sv_2025', t_rest())
        resp = policy.remove_from_storage(lun1, lun2)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_modify_policy(self):
        policy = UnityIoLimitPolicy('qp_4', t_rest())
        policy.modify(
            name="new_name",
            policy_type=IOLimitPolicyTypeEnum.DENSITY_BASED,
            max_iops_density=100, max_kbps_density=1024)


class UnityLinkAggregationTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        la = UnityLinkAggregation('spa_la_2', cli=t_rest())
        assert_that(la.is_link_up, equal_to(False))
        assert_that(la.mac_address, equal_to('00:60:16:5C:08:E1'))
        assert_that(la.master_port.get_id(), "spa_eth2")
        assert_that(la.mtu, equal_to(9000))
        assert_that(
            la.parent_storage_processor,
            equal_to(UnityStorageProcessor.get(t_rest(), 'spa')))
        assert_that(len(la.ports), equal_to(2))
        assert_that(la.supported_mtus, only_contains(1500, 9000))

    @patch_rest
    def test_create(self):
        eth_2 = UnityEthernetPort.get(t_rest(), 'spa_eth2')
        eth_3 = UnityEthernetPort.get(t_rest(), 'spa_eth3')
        la = UnityLinkAggregation.create(t_rest(), [eth_2, eth_3], 9000)
        assert_that(la.get_id(), equal_to('spa_la_2'))

    @patch_rest
    def test_create_already_exist(self):
        def do():
            eth_2 = UnityEthernetPort.get(t_rest(), 'spa_eth2')
            eth_4 = UnityEthernetPort.get(t_rest(), 'spa_eth4')
            UnityLinkAggregation.create(t_rest(), [eth_2, eth_4], 1500)

        assert_that(do, raises(UnityEthernetPortAlreadyAggregatedError))

    @patch_rest
    def test_modify(self):
        la = UnityLinkAggregation.get(t_rest(), 'spa_la_2')
        la.modify(mtu=1500,
                  remove_ports=[UnityEthernetPort.get(t_rest(), "spa_eth2")],
                  add_ports=[UnityEthernetPort.get(t_rest(), "spa_eth4")])

    @patch_rest()
    def test_list_la_unsupported(self):
        def f():
            UnityLinkAggregation.get(t_rest("4.0.0"))

        assert_that(f, raises(SystemAPINotSupported))

    @patch_rest(output='link_aggregation_not_supported.json')
    def test_link_aggregation_not_supported(self):
        ports = UnityLinkAggregationList(cli=t_rest("4.1.2"))
        assert_that(len(ports), equal_to(0))


class UnityIscsiNodeTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        node = UnityIscsiNode(_id='iscsinode_spb_eth3', cli=t_rest())
        assert_that(node.id, equal_to('iscsinode_spb_eth3'))
        assert_that(node.name, equal_to('iqn.1992-04.com.emc:cx.'
                                        'fnm00150600267.b1'))
        assert_that(node.alias, equal_to('0267.b1'))

    @patch_rest
    def test_get_nested_properties(self):
        node = UnityIscsiNode(_id='iscsinode_spb_eth3', cli=t_rest())
        assert_that(node.ethernet_port.id, equal_to('spb_eth3'))
        assert_that(node.ethernet_port.connector_type,
                    equal_to(ConnectorTypeEnum.RJ45))
        assert_that(node.ethernet_port.speed,
                    equal_to(EPSpeedValuesEnum._10GbPS))
        assert_that(node.ethernet_port.supported_speeds,
                    only_contains(
                        EPSpeedValuesEnum.AUTO,
                        EPSpeedValuesEnum._100MbPS,
                        EPSpeedValuesEnum._1GbPS,
                        EPSpeedValuesEnum._10GbPS))
        assert_that(node.ethernet_port.health, instance_of(UnityHealth))


class UnitySasPortTest(TestCase):
    @patch_rest
    def test_get_properties(self):
        port = UnitySasPort(_id='spa_sas4', cli=t_rest())
        assert_that(port.id, equal_to('spa_sas4'))
        assert_that(port.current_speed, equal_to(SpeedValuesEnum._12Gbps))
        assert_that(port.connector_type,
                    equal_to(ConnectorTypeEnum.MINI_SAS_HD))
        assert_that(port.health, instance_of(UnityHealth))
        assert_that(port.needs_replacement, equal_to(False))
        assert_that(port.name, equal_to('SP A SAS Port 4'))
        assert_that(port.port, equal_to(0))
        assert_that(port.storage_processor.id, equal_to('spa'))

    @patch_rest
    def test_get_nested_properties(self):
        port = UnitySasPort(_id='spa_sas4', cli=t_rest())
        assert_that(port.parent_storage_processor.id, equal_to('spa'))
        assert_that(port.parent_storage_processor.name, equal_to('SP A'))
        assert_that(port.parent_io_module.id, equal_to('ioa'))
        assert_that(port.parent_io_module.name, equal_to('IO_Module_A'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import pytest
from storops_comptest import t_vnx
from hamcrest import assert_that, equal_to, only_contains, none

__author__ = 'Cedric Zhuang'


vnx = t_vnx()


@pytest.mark.skipif(not vnx.is_snap_enabled(),
                    reason='snap feature not available')
def test_snap_create(vnx_gf):
    assert_that(vnx_gf.snap.existed, equal_to(True))
    assert_that(vnx_gf.snap.source_luns, only_contains(vnx_gf.lun.lun_id))
    assert_that(vnx_gf.snap.primary_luns, only_contains(vnx_gf.lun.lun_id))


@pytest.mark.skipif(not vnx.is_snap_enabled(),
                    reason='snap feature not available')
def test_attach_detach_snap(vnx_gf):
    smp_name = vnx_gf.add_lun_name()
    smp = vnx_gf.lun.create_mount_point(name=smp_name)
    smp.attach_snap(vnx_gf.snap)
    smp.update()
    assert_that(smp.attached_snapshot.name, equal_to(vnx_gf.snap.name))
    smp.detach_snap()
    smp.update()
    assert_that(smp.attached_snapshot, none())


@pytest.mark.skipif(not vnx.is_snap_enabled(),
                    reason='snap feature not available')
def test_delete_snap(vnx_gf):
    snap_name = vnx_gf.add_snap_name()
    vnx_gf.lun.create_snap(snap_name)
    snap = vnx_gf.vnx.get_snap(name=snap_name)
    snap.delete()
    snap.update()
    assert_that(snap.existed, equal_to(False))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import re
import six
from xml import etree

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


class XMLAPIParser(object):
    def __init__(self):
        # The following Boolean acts as the flag for the common sub-element.
        # For instance:
        #     <CifsServers>
        #         <li> server_1 </li>
        #     </CifsServers>
        #     <Alias>
        #         <li> interface_1 </li>
        #     </Alias>
        self.tag = None

        self.elt = {}
        self.stack = []

    @staticmethod
    def _delete_ns(tag):
        i = tag.find('}')
        if i >= 0:
            tag = tag[i + 1:]
        return tag

    def parse(self, xml):
        result = {
            'type': None,
            'taskId': None,
            'maxSeverity': None,
            'objects': [],
            'problems': [],
        }

        events = ("start", "end")

        context = etree.ElementTree.iterparse(six.BytesIO(xml.encode('utf-8')),
                                              events=events)
        for action, elem in context:
            self.tag = self._delete_ns(elem.tag)

            func = self._get_func(action, self.tag)
            self.track_stack(action, elem)

            if func in vars(XMLAPIParser):
                if action == 'start':
                    eval('self.' + func)(elem, result)
                elif action == 'end':
                    eval('self.' + func)(elem, result)

        return result

    def track_stack(self, action, elem):
        if action == 'start':
            self.stack.append(elem)
        elif action == 'end':
            self.stack.pop()

    @staticmethod
    def _get_func(action, tag):
        if tag == 'W2KServerData':
            return action + '_' + 'w2k_server_data'

        temp_list = re.sub(r"([A-Z])", r" \1", tag).split()
        if temp_list:
            func_name = action + '_' + '_'.join(temp_list)
        else:
            func_name = action + '_' + tag
        return func_name.lower()

    @staticmethod
    def _copy_property(source, target):
        for key in source:
            target[key] = source[key]

    @classmethod
    def _append_elm_property(cls, elm, result, identifier):
        for obj in result['objects']:
            if cls.has_identifier(obj, elm, identifier):
                for key, value in elm.attrib.items():
                    obj[key] = value

    @staticmethod
    def has_identifier(obj, elm, identifier):
        return (identifier in obj and
                identifier in elm.attrib and
                elm.attrib[identifier] == obj[identifier])

    def _append_element(self, elm, result, identifier):
        sub_elm = {}
        self._copy_property(elm.attrib, sub_elm)

        for obj in result['objects']:
            if self.has_identifier(obj, elm, identifier):
                if self.tag in obj:
                    obj[self.tag].append(sub_elm)
                else:
                    obj[self.tag] = [sub_elm]

    def start_task_response(self, elm, result):
        result['type'] = 'TaskResponse'
        self._copy_property(elm.attrib, result)

    @staticmethod
    def start_fault(_, result):
        result['type'] = 'Fault'

    def _parent_tag(self):
        if len(self.stack) >= 2:
            parent = self.stack[-2]
            ret = self._delete_ns(parent.tag)
        else:
            ret = None
        return ret

    def start_status(self, elm, result):
        parent_tag = self._parent_tag()
        if parent_tag == 'TaskResponse':
            result['maxSeverity'] = elm.attrib['maxSeverity']
        elif parent_tag in ['MoverStatus', 'Vdm', 'MoverHost']:
            self.elt['maxSeverity'] = elm.attrib['maxSeverity']

    def start_query_status(self, elm, result):
        result['type'] = 'QueryStatus'
        self._copy_property(elm.attrib, result)

    def start_problem(self, elm, result):
        self.elt = {}
        self._copy_property(elm.attrib, self.elt)
        result['problems'].append(self.elt)

    def start_description(self, elm, _):
        self.elt['Description'] = elm.text

    def start_action(self, elm, _):
        self.elt['Action'] = elm.text

    def start_diagnostics(self, elm, _):
        self.elt['Diagnostics'] = elm.text

    def start_file_system(self, elm, result):
        self._as_object(elm, result)

    def start_file_system_capacity_info(self, elm, result):
        identifier = 'fileSystem'

        self._append_elm_property(elm, result, identifier)

    def start_storage_pool(self, elm, result):
        self._as_object(elm, result)

    def start_system_storage_pool_data(self, elm, _):
        self._copy_property(elm.attrib, self.elt)

    def start_mover(self, elm, result):
        self._as_object(elm, result)

    def start_mover_host(self, elm, result):
        self._as_object(elm, result)

    def start_nfs_export(self, elm, result):
        self._as_object(elm, result)

    def _as_object(self, elm, result):
        self.elt = {}
        self._copy_property(elm.attrib, self.elt)
        result['objects'].append(self.elt)

    def start_mover_status(self, elm, result):
        identifier = 'mover'
        self._append_elm_property(elm, result, identifier)

    def start_mover_route(self, elm, result):
        self._append_element(elm, result, 'mover')

    def start_mover_deduplication_settings(self, elm, result):
        self._append_element(elm, result, 'mover')

    def start_mover_dns_domain(self, elm, result):
        self._append_element(elm, result, 'mover')

    def start_mover_interface(self, elm, result):
        self._append_element(elm, result, 'mover')

    def start_logical_network_device(self, elm, result):
        self._append_element(elm, result, 'mover')

    def start_vdm(self, elm, result):
        self._as_object(elm, result)

    def _add_element(self, name, item):
        if name not in self.elt:
            self.elt[name] = []
        self.elt[name].append(item)

    def start_li(self, elm, _):
        parent_tag = self._parent_tag()
        host_nodes = ('AccessHosts', 'RwHosts', 'RoHosts', 'RootHosts')
        if parent_tag == 'CifsServers':
            self._add_element('CifsServers', elm.text)
        elif parent_tag == 'Aliases':
            self._add_element('Aliases', elm.text)
        elif parent_tag == 'Interfaces':
            self._add_element('Interfaces', elm.text)
        elif parent_tag in host_nodes:
            if parent_tag not in self.elt:
                self.elt[parent_tag] = []
            self.elt[parent_tag].append(elm.text)

    def start_cifs_server(self, elm, result):
        self._as_object(elm, result)

    def start_w2k_server_data(self, elm, _):
        self._copy_property(elm.attrib, self.elt)

    def start_cifs_share(self, elm, result):
        self._as_object(elm, result)

    def start_checkpoint(self, elm, result):
        self._as_object(elm, result)

    def start_ro_file_system_hosts(self, elm, _):
        self._copy_property(elm.attrib, self.elt)

    def start_standalone_server_data(self, elm, _):
        self._copy_property(elm.attrib, self.elt)

    def start_fibre_channel_device_data(self, elm, _):
        self._copy_attrib_to_parent(elm)

    def start_network_device_data(self, elm, _):
        self._copy_attrib_to_parent(elm)

    def _copy_attrib_to_parent(self, elm):
        if len(self.stack) >= 2:
            parent = self.stack[-2]
            for k, v in elm.attrib.items():
                parent.attrib[k] = v

    def start_mover_motherboard(self, elm, result):
        self._append_element(elm, result, 'moverHost')

    def end_physical_device(self, elm, result):
        self._append_element(elm, result, 'moverHost')

    def start_fc_descriptor(self, elm, result):
        self._append_element(elm, result, 'moverHost')

    def start_mount(self, elm, result):
        self._as_object(elm, result)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import os
from os import path
from unittest import TestCase

from hamcrest import assert_that, equal_to, instance_of, only_contains, \
    raises, contains_string, greater_than, is_not

from storops.lib.common import get_file_size, get_local_folder
from storops.unity.enums import NodeEnum
from storops.unity.resource.health import UnityHealth
from storops.unity.resource.sp import UnityStorageProcessor, \
    UnityStorageProcessorList
from storops.unity.resource.system import UnityDpe, UnitySystem
from storops_test.unity.rest_mock import t_rest, patch_rest, t_unity

__author__ = 'Cedric Zhuang'


class UnityStorageProcessorTest(TestCase):

    @property
    @patch_rest
    def sp_list(self):
        return t_unity().get_sp()

    @patch_rest
    def test_properties(self):
        sp = UnityStorageProcessor(_id='spa', cli=t_rest())
        assert_that(sp.id, equal_to('spa'))
        assert_that(sp.existed, equal_to(True))
        assert_that(sp.health, instance_of(UnityHealth))
        assert_that(sp.needs_replacement, equal_to(False))
        assert_that(sp.is_rescue_mode, equal_to(False))
        assert_that(sp.model, equal_to('OBERON CANISTER 10C 105W 2.6G'))
        assert_that(sp.slot_number, equal_to(0))
        assert_that(sp.name, equal_to('SP A'))
        assert_that(sp.emc_part_number, equal_to('110-297-008C-04'))
        assert_that(sp.emc_serial_number, equal_to('CF2HF150300001'))
        assert_that(sp.manufacturer, equal_to(''))
        assert_that(sp.vendor_part_number, equal_to(''))
        assert_that(sp.vendor_serial_number, equal_to(''))
        assert_that(sp.sas_expander_version, equal_to('2.7.1'))
        assert_that(sp.bios_firmware_revision, equal_to('30.89'))
        assert_that(sp.post_firmware_revision, equal_to('21.2'))
        assert_that(sp.memory_size, equal_to(65536))
        assert_that(sp.parent_dpe, instance_of(UnityDpe))

    @patch_rest
    def test_sp_to_node_enum(self):
        sp = UnityStorageProcessor(_id='spa', cli=t_rest())
        assert_that(sp.to_node_enum(), equal_to(NodeEnum.SPA))
        sp = UnityStorageProcessor(_id='spb', cli=t_rest())
        assert_that(sp.to_node_enum(), equal_to(NodeEnum.SPB))
        sp = UnityStorageProcessor(_id='wrong', cli=t_rest())
        assert_that(sp.to_node_enum(), equal_to(NodeEnum.UNKNOWN))

    @patch_rest
    def test_get_all_and_property_query(self):
        sp_list = UnityStorageProcessorList(cli=t_rest())
        assert_that(len(sp_list), equal_to(2))
        assert_that(sp_list.name, only_contains('SP A', 'SP B'))

    @patch_rest
    def test_get_all_and_property_not_found(self):
        def f():
            sp_list = UnityStorageProcessorList(cli=t_rest())
            return sp_list.not_found

        assert_that(f, raises(AttributeError, 'not_found'))

    @patch_rest
    def test_metric_utilization(self):
        spa, spb = self.sp_list
        assert_that(spa.utilization, equal_to(22))
        assert_that(spb.utilization, equal_to(33))

    @patch_rest
    def test_metric_nfs_write_mbps(self):
        spa, spb = self.sp_list
        assert_that(spa.nfs_write_mbps, equal_to(3.9))
        assert_that(spb.nfs_write_mbps, equal_to(4.1))

    @patch_rest
    def test_metric_nfs_read_mbps(self):
        spa, spb = self.sp_list
        assert_that(spa.nfs_read_mbps, equal_to(3.7))
        assert_that(spb.nfs_read_mbps, equal_to(3.8))

    @patch_rest
    def test_metric_nfs_write_iops(self):
        spa, spb = self.sp_list
        assert_that(spa.nfs_write_iops, equal_to(3.5))
        assert_that(spb.nfs_write_iops, equal_to(3.6))

    @patch_rest
    def test_metric_nfs_read_iops(self):
        spa, spb = self.sp_list
        assert_that(spa.nfs_read_iops, equal_to(3.3))
        assert_that(spb.nfs_read_iops, equal_to(3.4))

    @patch_rest
    def test_metric_cifs_write_mbps(self):
        spa, spb = self.sp_list
        assert_that(spa.cifs_write_mbps, equal_to(3.1))
        assert_that(spb.cifs_write_mbps, equal_to(3.2))

    @patch_rest
    def test_metric_cifs_read_mbps(self):
        spa, spb = self.sp_list
        assert_that(spa.cifs_read_mbps, equal_to(2.8))
        assert_that(spb.cifs_read_mbps, equal_to(2.9))

    @patch_rest
    def test_metric_cifs_write_iops(self):
        spa, spb = self.sp_list
        assert_that(spa.cifs_write_iops, equal_to(2.6))
        assert_that(spb.cifs_write_iops, equal_to(2.7))

    @patch_rest
    def test_metric_cifs_read_iops(self):
        spa, spb = self.sp_list
        assert_that(spa.cifs_read_iops, equal_to(2.4))
        assert_that(spb.cifs_read_iops, equal_to(2.5))

    @patch_rest
    def test_metric_block_write_mbps(self):
        spa, spb = self.sp_list
        assert_that(spa.block_write_mbps, equal_to(30))
        assert_that(spb.block_write_mbps, equal_to(40))

    @patch_rest
    def test_metric_block_read_mbps(self):
        spa, spb = self.sp_list
        assert_that(spa.block_read_mbps, equal_to(1.9))
        assert_that(spb.block_read_mbps, equal_to(2.1))

    @patch_rest
    def test_metric_block_write_iops(self):
        spa, spb = self.sp_list
        assert_that(spa.block_write_iops, equal_to(1.7))
        assert_that(spb.block_write_iops, equal_to(1.8))

    @patch_rest
    def test_metric_block_read_iops(self):
        spa, spb = self.sp_list
        assert_that(spa.block_read_iops, equal_to(1.5))
        assert_that(spb.block_read_iops, equal_to(1.6))

    @patch_rest
    def test_metric_temperature(self):
        spa, spb = self.sp_list
        assert_that(spa.temperature, equal_to(27))
        assert_that(spb.temperature, equal_to(28))

    @patch_rest
    def test_metric_core_count(self):
        spa, spb = self.sp_list
        assert_that(spa.core_count, equal_to(10))
        assert_that(spa.core_count, equal_to(10))

    @patch_rest
    def test_metric_net_out_mbps(self):
        spa, spb = self.sp_list
        assert_that(spa.net_out_mbps, equal_to(1.3))
        assert_that(spb.net_out_mbps, equal_to(1.4))

    @patch_rest
    def test_metric_net_in_mbps(self):
        spa, spb = self.sp_list
        assert_that(spa.net_in_mbps, equal_to(1.1))
        assert_that(spb.net_in_mbps, equal_to(1.2))

    @patch_rest
    def test_metric_block_cache_read_hit_ratio(self):
        spa, spb = self.sp_list
        assert_that(spa.block_cache_read_hit_ratio, equal_to(87.0))
        assert_that(spb.block_cache_read_hit_ratio, equal_to(88.0))

    @patch_rest
    def test_metric_block_cache_write_hit_ratio(self):
        spa, spb = self.sp_list
        assert_that(spa.block_cache_write_hit_ratio, equal_to(89.0))
        assert_that(spb.block_cache_write_hit_ratio, equal_to(90.0))

    @patch_rest
    def test_csv(self):
        csv = self.sp_list.get_metrics_csv()
        assert_that(csv, contains_string('timestamp,id,name,'))
        assert_that(csv, contains_string('block_read_iops,block_read_mbps'))
        assert_that(csv, contains_string('2016-11-21 09:10:00+00:00,spa'))
        assert_that(csv, contains_string('spa,SP A,89,87.0,89.0'))
        assert_that(csv, contains_string('spb,SP B,78,88.0,90.0'))

    FILENAME = path.join(get_local_folder(),
                         'unittest_sp_metric_persist_csv_file.csv')

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(cls.FILENAME):
            os.remove(cls.FILENAME)

    @patch_rest
    def test_persist_csv(self):
        self.sp_list.persist_metric_data(self.FILENAME)
        assert_that(os.path.exists(self.FILENAME), equal_to(True))
        file_size = get_file_size(self.FILENAME)

        self.sp_list.persist_metric_data(self.FILENAME)
        new_file_size = get_file_size(self.FILENAME)
        assert_that(new_file_size, greater_than(file_size))

    @patch_rest
    def test_repr_with_metric(self):
        spa, _ = self.sp_list
        assert_that(str(spa), contains_string('"nfs_write_mbps":'))

    @patch_rest
    def test_repr_without_metric(self):
        spa, _ = UnitySystem('10.244.223.61').get_sp()
        assert_that(str(spa), is_not(contains_string('"nfs_write_mbps":')))

    @patch_rest
    def test_default_metric_csv_filename(self):
        sp_list = UnitySystem('10.244.223.61').get_sp()
        filename = sp_list.get_default_metric_csv_filename()
        assert_that(filename, contains_string('.storops'))
        assert_that(filename,
                    contains_string('10.244.223.61_storageProcessor.csv'))



# coding=utf-8
# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import unicode_literals

from unittest import TestCase

import ddt
from hamcrest import assert_that, equal_to

from storops.unity.enums import FilesystemSnapAccessTypeEnum, \
    ScheduleTypeEnum, ScheduleVersionEnum, HostContainerTypeEnum, \
    VasaProviderStateEnum
from storops.unity.resource.host_container import UnityHostContainer, \
    UnityHostContainerList
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Ryan Liang'


@ddt.ddt
class UnityHostContainerTest(TestCase):
    @patch_rest
    def test_properties(self):
        container = UnityHostContainer(_id='mss_3', cli=t_rest())
        assert_that(container.existed, equal_to(True))
        # TODO(ryan)
        # assert_that(container.version, equal_to(ScheduleVersionEnum.LEGACY))
        # assert_that(container.name, equal_to('Snapshot Schedule 1'))
        # assert_that(container.is_default, equal_to(False))
        # assert_that(container.is_modified, equal_to(False))
        # assert_that(str(container.modification_time),
        #             equal_to('2020-03-17 07:37:56.695000+00:00'))
        # assert_that(len(container.rules), equal_to(1))
        # assert_that(container.rules[0].type,
        #             equal_to(ScheduleTypeEnum.DAY_AT_HHMM))
        # assert_that(len(container.rules[0].days_of_week), equal_to(0))
        # assert_that(container.rules[0].access_type,
        #             equal_to(FilesystemSnapAccessTypeEnum.CHECKPOINT))
        # assert_that(container.rules[0].minute, equal_to(0))
        # assert_that(container.rules[0].hours, equal_to([4]))
        # assert_that(container.rules[0].interval, equal_to(1))
        # assert_that(container.rules[0].is_auto_delete, equal_to(True))
        # assert_that(container.rules[0].retention_time, equal_to(0))
        # assert_that(container.is_sync_replicated, equal_to(False))
        # assert_that(len(container.luns), equal_to(1))
        # assert_that(container.luns[0].get_id(), equal_to('sv_2'))

    @patch_rest
    def test_get_all(self):
        containers = UnityHostContainerList(cli=t_rest())
        assert_that(containers[0].existed, equal_to(True))
        assert_that(len(containers), equal_to(1))

    @patch_rest
    def test_create(self):
        _type, name, _, _, hosts = UnityHostContainer.recommend(
            t_rest(), address='172.18.15.201',
            username='administrator@vsphere.local', password='Test123!'
        )
        container = UnityHostContainer.create(
            t_rest(), _type,
            name, name, 'administrator@vsphere.local',
            'Test123!', local_username='admin', local_password='Test123!',
            potential_hosts=hosts
        )
        assert_that(container.get_id(), equal_to('mss_3'))

    @patch_rest
    def test_recommend(self):
        _type, name, version, state, hosts = UnityHostContainer.recommend(
            t_rest(),
            address='172.18.15.201',
            username='administrator@vsphere.local',
            password='Test123!')
        assert_that(_type, equal_to(2))
        assert_that(name, equal_to('172.18.15.201'))
        assert_that(version, equal_to('7.0.0'))
        assert_that(state, equal_to(0))
        assert_that(len(hosts), equal_to(3))

    # TODO(ryan)
    # @patch_rest
    # def test_delete(self):
    #     container = UnityHostContainer(_id='mss_3', cli=t_rest())
    #     resp = container.delete()
    #     assert_that(resp.is_ok(), equal_to(True))
